"""piv2hdf: A package to convert PIV data to HDF5 files."""

import atexit
import pathlib
import shutil
import warnings
from typing import Dict

import h5rdmtoolbox as h5tbx
import yaml

from . import cache
from ._logger import logger, set_loglevel
from ._version import __version__
from .config import set_config, set_pivattrs, get_config, reset_pivattrs
from .convention import cv, CONVENTION_FILENAME
from .interface import PIVSnapshot, PIVPlane, PIVMultiPlane, UserDefinedHDF5Operation
from .user import USER_DIRECTORY

__all__ = ['PIVSnapshot', 'PIVPlane', 'PIVMultiPlane', 'logger', 'set_loglevel',
           'get_config', 'set_config', 'set_pivattrs', 'reset_pivattrs', 'UserDefinedHDF5Operation',
           '__version__', 'CONVENTION_FILENAME', 'cv']

h5tbx.use(None)

__this_dir__ = pathlib.Path(__file__).parent
_data_dir = __this_dir__ / 'data'


class Contacts:
    """A class to register contacts for the PIV convention."""

    def __init__(self):
        # read from user_dir
        self.contacts_filename = USER_DIRECTORY / 'contacts.yaml'
        if self.contacts_filename.exists():
            with open(self.contacts_filename, 'r') as f:
                self.registered_contacts = yaml.safe_load(f)
        else:
            self.registered_contacts = {}
        self.save()

    def __contains__(self, item):
        return item in self.registered_contacts

    def __getitem__(self, item):
        return self.registered_contacts[item]

    def __repr__(self):
        return f'Contacts({self.registered_contacts})'

    def __len__(self):
        return len(self.registered_contacts)

    def save(self):
        """Write the yaml file to user directory"""
        with open(self.contacts_filename, 'w') as f:
            yaml.safe_dump(self.registered_contacts, f)

    def register(self, **orcid: Dict):
        """Add a new contact. Provide one or multiple ORCID(s) as a dictionary."""
        self.registered_contacts.update(orcid)
        self.save()

    def clear(self):
        """removes all contacts"""
        self.registered_contacts = {}
        self.save()


contacts = Contacts()


@atexit.register
def clean_temp_data():
    """cleaning up the tmp directory"""
    for fname in cache.tmp_filenames:
        try:
            fname.unlink()
        except OSError:
            warnings.warn(f'Could not delete file {fname}. Consider deleting it manually.')
    for fdir in cache.tmp_dirnames:
        try:
            shutil.rmtree(fdir)
        except OSError:
            warnings.warn(f'Could not delete directory {fdir}. Consider deleting it manually.')



DAVIS_SOFTWARE = dict(name='Davis', version=999,
                      author='LaVision GmbH',
                      url='https://www.lavision.de')

DEFAULT_DATASET_LONG_NAMES = {'x': 'x-coordinate',
                              'y': 'y-coordinate',
                              'z': 'z-coordinate',
                              'ix': 'x-pixel-coordinate',
                              'iy': 'y-pixel-coordinate',
                              'reltime': 'reltime'}



import json
import os
import pathlib
import re
from datetime import datetime
from pathlib import Path
from typing import Dict, Tuple, Union, List, Optional

import h5py
import numpy as np
import xarray as xr
from netCDF4 import Dataset as ncDataset
from scipy.interpolate import LinearNDInterpolator
from scipy.spatial import Delaunay

from piv2hdf import logger
from . import parameter as pivview_parameter
from .const import *
from .. import flags
from ..config import get_config
from ..interface import PIVFile, UserDefinedHDF5Operation, PIVData
from ..time import create_recording_datetime_dataset
from ..utils import (is_time, get_uint_type,
                     parse_z)

__this_file__ = pathlib.Path(__file__).resolve()


def _find_common_entries(dictionaries):
    common_dict = {}

    # Iterate over the keys of the first dictionary
    for key in dictionaries[0]:
        # Check if all dictionaries contain the key
        if all(key in dictionary for dictionary in dictionaries):
            # Check if the values are the same for all dictionaries
            if all(d[key] == dictionaries[0][key] for d in dictionaries):
                # Check if the value is a nested dictionary
                if isinstance(dictionaries[0][key], dict):
                    # Recursively find common entries in nested dictionaries
                    nested_dicts = [d[key] for d in dictionaries]
                    common_dict[key] = _find_common_entries(nested_dicts)
                else:
                    common_dict[key] = dictionaries[0][key]

    return common_dict


def _process_pivview_root_attributes(root_attrs):
    """Create new or renames or deletes dictionary entries"""
    root_attrs['recording_date'] = root_attrs['creation_date']

    del root_attrs['creation_date']
    if 'filename' in root_attrs:
        del root_attrs['filename']
    if 'image_bkgd_file' in root_attrs:
        del root_attrs['image_bkgd_file']  # should be dataset!
    if 'image_bkgd2_file' in root_attrs:
        del root_attrs['image_bkgd2_file']  # should be dataset!
    if 'image_mask_file' in root_attrs:
        del root_attrs['image_mask_file']  # should be dataset or can always be reconstructed from piv_flags!
    return root_attrs


def process_pivview_nc_data(nc_file: pathlib.Path, interpolate: bool,
                            masking: str,
                            timestep: float, time_unit: str = 's',
                            z_source: str = 'coord_min',
                            compute_dwdz: bool = False,
                            build_coord_datasets: bool = True
                            ) -> Tuple[Dict, Dict, Dict]:
    """
    Reads data and attributes from netCDF file. Results are stored in dictionary. Interpolation
    to fill "holes"/masked areas is applied if asked. Data arrays x, y, z and time are created.
    Array shape is changed from (1, ny, nx) and (1, ny, nx, nv) to (ny, nx). Thus, new variables are created:
        velocity (1, ny, nx, 2[3]) --> u (ny, nx), v (ny, nx) [, w (ny, nx)]
        piv_data (1, ny, nx, 2[3]) --> dx (ny, nx), dy (ny, nx) [, dz (ny, nx)]
        piv_peak1 (1, ny, nx, 3) --> piv_peak1_dx (ny, nx), piv_peak1_dy (ny, nx), piv_peak1_corr
        piv_peak2 (see above)
        piv_peak3 (see above)

    Parameters
    ----------
    nc_file : Path
        path to nc file
    interpolate : bool, optional=True
        Use space interpolation (linear) in each timeframe to patch holes
        left out after masking.
    masking : str, optional='sepeaks'
        Masking mode from piv flags. Can be slack or sepeaks or strict.
            strict: only uses true good values
            slack: also uses interpolated values.
            sepeaks: uses strict and second corr. peaks.
        The default is 'sepeaks'.
    timestep : float
        Time step in [s] relative to measurement start. If astropy-quantity
        is passed, unit is taken from there.
    time_unit : str, optional='s'
        unit taken for time if time was not of type astropy.quantity.
    z_source: str or tuple, optional='coord_min'
        The z-location can be specified manually by passing an astropy quantity
        or a list/tuple in form  (value: float, unit: str).
        It is also possible to detect the z-location from the nc-file (experience shows
        that this was not always correctly set by the experimenter...). Alternative,
        the z-location is tried to determine by the file name.
        Automatic detection from nc file can be used scanning one of the following
        attributes which then are to be passed as string for this parameter:
            * 'coord_min'
            * 'coord_max'
            * 'origin_offset'
            * 'file'
    compute_dwdz: bool, optional=False
        Computing the out-of-plane gradient dwdz evaluating the continuity equation.
        Only valid for incompressible flows!!!
    build_coord_datasets: bool, optional=True
        Whether to generate x,y,z,t datasets. For single snapshot HDF built this should
        be set True. For plane or case generation except the first snapshot, this can be
        set to False and thus reduce computation time.

    Returns
    -------
    piv_data_array_dict : dict
        Dictionary containing the arrays
    nc_root_attributes : dict
        Attribute dictionary of root variables
    variable_attributes : dict
        Attribute dictionary of dataset variables

    Notes
    -----
    Credits to M. Elfner (ITS, Karlsruhe Institute of Technology)

    """

    # TODO get rid of nc_root_attributes

    def _build_meshgrid_xy(coord_min, coord_max, width, height):
        """generates coord meshgrid from velocity attribute stating min/max of coordinates x, y"""
        _x = np.linspace(coord_min[0], coord_max[0], width)
        _y = np.linspace(coord_min[1], coord_max[1], height)
        return np.meshgrid(_x, _y)

    # read netCDF4 data (array, root attributes and variable attributes)
    with ncDataset(nc_file, "r") as nc_rootgrp:
        dims = nc_rootgrp.dimensions
        d, h, w, vd = (dims['data_array_depth'].size, dims['data_array_height'].size,
                       dims['data_array_width'].size, dims['vec_dims'].size)

        if d > 1:
            logger.critical('File with depth not implemented')
            raise NotImplementedError('File with depth not implemented')

        root_attributes = {key: nc_rootgrp.getncattr(key) for key in
                           ('file_content', 'creation_date', 'software')}
        nc_root_attributes = _process_pivview_root_attributes(
            {attr: nc_rootgrp.getncattr(attr) for attr in nc_rootgrp.ncattrs()})

        variable_attributes = {}

        # Variable information.
        nc_data_array_dict = nc_rootgrp.variables

        if nc_root_attributes['outlier_interpolate'] and masking != 'slack':
            logger.debug('Outlier masking does not conform with pivview settings in nc '
                         f'(outlier_interpolate={nc_root_attributes["outlier_interpolate"]} vs {masking}) - '
                         f'averages might differ')
        elif nc_root_attributes['outlier_try_other_peak'] and masking != 'sepeaks':
            logger.debug('Outlier masking does not conform with pivview settings in nc '
                         f'(outlier_interpolate={nc_root_attributes["outlier_interpolate"]} vs {masking}) - '
                         f'averages might differ')

        # processed
        piv_data_array_dict = {}

        for v in nc_data_array_dict.keys():
            variable_attributes[v] = {key: nc_data_array_dict[v].getncattr(key) for key in
                                      nc_data_array_dict[v].ncattrs()}

        if 'piv_flags' in nc_data_array_dict:
            pivflags = np.asarray(nc_data_array_dict['piv_flags'])[0, ...]  # dtype int8

            # optimize memory usage
            if pivflags.dtype == 'int8':
                dtype_pivflags = 'uint8'
            elif pivflags.dtype == 'int16':
                dtype_pivflags = 'uint16'
            elif pivflags.dtype == 'int32':
                dtype_pivflags = 'uint32'
            else:
                raise ValueError(f'Invalid dtype for piv_flags: {pivflags.dtype}')
            piv_data_array_dict['piv_flags'] = pivflags.astype(dtype_pivflags)
        else:
            logger.warning(f'PIV flags not found, masking disabled, mode {masking} not used')
            mask = np.ones((h, w)).astype(bool)

        for k, v in nc_data_array_dict.items():
            if k == 'piv_data':
                for i, name in zip(range(v.shape[-1]), ('dx', 'dy', 'dz')):
                    data = v[0, :, :, i]
                    piv_data_array_dict[name] = data
                    variable_attributes[name] = {'units': 'pixel'}
            elif k == 'velocity':
                for i, name in zip(range(v.shape[-1]), ('u', 'v', 'w')):
                    data = v[0, :, :, i]
                    piv_data_array_dict[name] = data
                    variable_attributes[name] = {'units': variable_attributes['velocity']['units']}
            elif k == 'piv_flags':
                variable_attributes[k].update({'units': ''})  # , 'standard_name': 'piv_flags'
                continue  # We take the flags from above to save computation time
            elif k in ('piv_peak1', 'piv_peak2', 'piv_peak3'):
                for i, suffix, units in zip(range(v.shape[-1]),
                                            ('_dx', '_dy', '_corr'),
                                            ('pixel', 'pixel', '')):
                    data = v[0, :, :, i]
                    name = f'{k}{suffix}'
                    piv_data_array_dict[name] = data
                    variable_attributes[name] = {'units': units}
            else:
                data = v[0, ...]
                piv_data_array_dict[k] = data
                if k == 'piv_correlation_coeff':
                    variable_attributes[k].update({'units': ''})
                elif k == 'piv_snr_data':
                    variable_attributes[k].update({'units': ''})
                elif k == 'piv_3c_residuals':
                    variable_attributes[k].update({'units': '',
                                                   'long_name': 'least square residual for z_velocity',
                                                   'comment': 'Residuals from least-squares fit to determined '
                                                              'out-of-plane component. It is a measure of quality of '
                                                              'the vector reconstruction and should be lower than 0.5 '
                                                              'pixel'})
                elif 'velocity gradient' in variable_attributes[k]['long_name']:
                    variable_attributes[k].update({'units': f"1/{variable_attributes['velocity']['units'][-1]}"})

        if interpolate:
            # While there exist some base functions assuming reg. grids and / or
            # using splines, a true interpolation considering distances is the way
            # to go here. Any other, faster methods rely on splines which are not
            # bounded!
            # Qhull on location with valid data
            if 'x' and 'y' not in piv_data_array_dict.keys():
                xm, ym = _build_meshgrid_xy(nc_data_array_dict['velocity'].coord_min,
                                            nc_data_array_dict['velocity'].coord_max,
                                            w, h)
            xv = xm.ravel()[mask.ravel()]
            yv = ym.ravel()[mask.ravel()]
            xi = xm.ravel()[~mask.ravel()]
            yi = ym.ravel()[~mask.ravel()]
            tri = Delaunay(np.stack((xv, yv)).T)

            # interpolate, create and evaluate, write
            for k, v in piv_data_array_dict.items():

                # skip iteration if k in the following variables
                if k in ('x', 'y', 'valid', 'piv_flags'):
                    continue

                data = v.copy()
                interpolator = LinearNDInterpolator(tri, data.ravel()[mask.ravel()])
                interpolated_result = interpolator(xi, yi)
                data[~mask] = interpolated_result

                piv_data_array_dict[k] = data

        # Check if source for w gradients is available, if so compute.
        # CAREFUL with indexing: Meshgrids are YX indexed by default!
        if 'w' in piv_data_array_dict or build_coord_datasets:
            fr, to = nc_data_array_dict['velocity'].coord_min, nc_data_array_dict['velocity'].coord_max
            px_fr, px_to = nc_data_array_dict['piv_data'].coord_min, nc_data_array_dict['piv_data'].coord_max

        if 'w' in piv_data_array_dict:
            # grid spacing is assumed to be homogeneous - is the case for PIV measurements!
            dx, dy = (to[:2] - fr[:2]) / (np.array([w, h]) - 1)

            piv_data_array_dict['dwdx'] = np.gradient(piv_data_array_dict['w'], dx, axis=1)
            piv_data_array_dict['dwdy'] = np.gradient(piv_data_array_dict['w'], dy, axis=0)

            _gradient_unit = f"1/{variable_attributes['velocity']['units'][-1]}"
            variable_attributes['dwdx'] = {'units': _gradient_unit,
                                           'long_name': 'velocity gradient dw/dx'}
            variable_attributes['dwdy'] = {'units': _gradient_unit,
                                           'long_name': 'velocity gradient dw/dy'}

            if compute_dwdz:
                if 'dudx' in piv_data_array_dict.keys() and 'dvdy' in piv_data_array_dict.keys():
                    logger.info("Gradient \"dwdz\" calculated from continuity equation assuming incompressible flow!")
                    piv_data_array_dict['dwdz'] = -piv_data_array_dict['dudx'] - piv_data_array_dict['dvdy']
                    variable_attributes['dwdz'] = {
                        'units': _gradient_unit,
                        'long_name': 'velocity gradient dw/dxz assuming incompressible flow',
                        # 'standard_name': 'z_derivative_of_w_velocity_assuming_incompressibility',
                        'comment': 'Calculated from continuity equation using dud and dvdy and assuming '
                                   'incompressible flow!'
                    }
                else:
                    logger.error(
                        "Could not compute dwdz based on continuity as du/dx and dv/dy are missing. Continuing ..."
                    )

        piv_data_array_dict['reltime'] = timestep
        variable_attributes['reltime'] = {'long_name': 'Recording time since start.',
                                          'units': time_unit}
        if not is_time(variable_attributes['reltime']['units']):
            raise AttributeError(f'Time unit is incorrect: {variable_attributes["t"]["unit"]}')

        # x,y,z,t are not part of PIVview netCDF variables
        if build_coord_datasets:  # can speed up computation when plane or case HDF files are generated
            # the velocity dataset has the attribute coord_min and coord_max from which the coordinates can be derived:
            piv_data_array_dict['x'] = np.linspace(fr[0], to[0], w)
            piv_data_array_dict['y'] = np.linspace(fr[1], to[1], h)
            variable_attributes['x'] = {'units': nc_root_attributes['length_conversion_units']}
            variable_attributes['y'] = {'units': nc_root_attributes['length_conversion_units']}
            if not px_fr[0] >= 0:
                raise ValueError(f'Invalid pixel coordinate: {px_fr[0]}')
            if not px_fr[1] >= 0:
                raise ValueError(f'Invalid pixel coordinate: {px_fr[1]}')

            piv_data_array_dict['ix'] = np.linspace(px_fr[0], px_to[0], w).astype(get_uint_type(px_to[0]))
            piv_data_array_dict['iy'] = np.linspace(px_fr[1], px_to[1], h).astype(get_uint_type(px_to[0]))
            variable_attributes['ix'] = {'long_name': 'pixel x-location of vector',
                                         'units': 'pixel'}
            variable_attributes['iy'] = {'long_name': 'pixel y-location of vector',
                                         'units': 'pixel'}

            # Z position information
            z_unit = nc_root_attributes['length_conversion_units']  # default unit of z comes from file
            if isinstance(z_source, str):
                if z_source == 'coord_min':
                    z = fr[2]
                elif z_source == 'coord_max':
                    z = to[2]
                elif z_source == 'origin_offset':
                    z = nc_root_attributes['origin_offset_z']
                elif z_source == 'file':
                    try:
                        z = float(
                            re.findall(
                                r'([-+\d].*)', os.path.split(os.path.dirname(nc_file))[-1])[0])
                    except Exception as e:
                        logger.warning(f'Z level detection failed due to: {e}')
                        z = 0
            elif isinstance(z_source, xr.DataArray):
                z = z_source.values
                z_unit = z_source.units
            elif isinstance(z_source, tuple) or isinstance(z_source, list):
                z, z_unit = z_source
            else:
                logger.warning('Z level detection failed: Invalid mode')
                z = 0

            piv_data_array_dict['z'] = z
            variable_attributes['z'] = {'units': z_unit}

        # correct shape if needed to (nz, nt, ny, nx) or (nz, nt, ny, nx, nv) respectively
        for k, v in piv_data_array_dict.items():
            if k not in ('x', 'y', 'z', 'reltime'):
                if v.ndim > 2:  # single var dataset or vector dataset
                    piv_data_array_dict[k] = v[0, ...]
                else:  # mask, coordinates
                    piv_data_array_dict[k] = v

    return piv_data_array_dict, root_attributes, variable_attributes


class PIVViewNcFile(PIVFile):
    """Interface class to a PIVview netCDF4 file for a 2D recording (using .par parameter files)

    Parameters
    ----------
    filename : pathlib.Path
        path to the nc file
    parameter : Union[None, pivview_parameter.PIVviewParamFile, pathlib.Path], optional
        path to the parameter file, by default None
    user_defined_hdf5_operations : optional UserDefinedHDF5Operation or List[UserDefinedHDF5Operation]
        Injects code that is executed after the HDF5 file is created, by default None
    kwargs : dict
        additional keyword arguments
    """
    suffix: str = '.nc'
    __parameter_cls__ = pivview_parameter.PIVviewParamFile

    def __init__(self,
                 filename: pathlib.Path,
                 parameter: Optional[Union[pivview_parameter.PIVviewParamFile, pathlib.Path]] = None,
                 user_defined_hdf5_operations: Optional[
                     Union[UserDefinedHDF5Operation, List[UserDefinedHDF5Operation]]] = None,
                 **kwargs):
        # if user_defined_hdf5_operations is None:
        #     user_defined_hdf5_operations = add_standard_name_operation
        super().__init__(filename, parameter, user_defined_hdf5_operations=user_defined_hdf5_operations, **kwargs)

    def read(self,
             relative_time: float,
             **kwargs) -> PIVData:
        """reads and processes nc data

        Parameters
        ----------
        relative_time : float
            relative time of the snapshot read
        kwargs:
            build_coord_datasets : bool, optional=True
                whether to build coordinate datasets, by default True

        Returns
        -------
        nc_data : dict
            dictionary of data arrays
        nc_root_attr : dict
            dictionary of root attributes
        nc_variable_attr : dict
            dictionary of variable attributes
        """

        build_coord_datasets = kwargs.get("build_coord_datasets", True)
        from . import config as pivview_config
        masking = pivview_config['masking']
        interpolation = pivview_config['interpolation']
        z_source = pivview_config['z_source']
        nc_data, nc_root_attr, nc_variable_attr = process_pivview_nc_data(self.filename,
                                                                          z_source=z_source,
                                                                          timestep=relative_time,
                                                                          masking=masking,
                                                                          interpolate=interpolation,
                                                                          build_coord_datasets=build_coord_datasets)
        # nc_root_attr['filename'] = nc_root_attr.pop('long_name')
        # unique_flags = np.unique(piv_data.data['piv_flags'][:])
        nc_variable_attr['piv_flags']['flag_meaning'] = json.dumps({flag.value: flag.name for flag in flags.Flags})
        return PIVData(nc_data, nc_variable_attr, nc_root_attr)

    def to_hdf(self,
               hdf_filename: pathlib.Path,
               relative_time: float,
               recording_dtime: Union[datetime, List[datetime]],  # iso-format of datetime!
               z: Union[str, float, None] = None,
               **kwargs) -> pathlib.Path:
        """converts the snapshot into an HDF file"""
        piv_data = self.read(relative_time=relative_time)
        if z is not None:
            piv_data.data['z'], piv_data.data_attrs['z']['units'] = parse_z(z)
        ny, nx = piv_data.data['y'].size, piv_data.data['x'].size
        # building HDF file
        if hdf_filename is None:
            _hdf_filename = Path.joinpath(self.filename.parent, f'{self.filename.stem}.hdf')
        else:
            _hdf_filename = hdf_filename
        with h5py.File(_hdf_filename, "w") as main:
            main.attrs['plane_directory'] = str(self.filename.parent.resolve())
            main.attrs['software'] = json.dumps(PIVVIEW_SOFTWARE)
            main.attrs['title'] = 'piv snapshot data'

            # process piv_parameters. there must be a parameter file at the parent location
            # piv_param_grp = main.create_group(PIV_PARAMETER_GRP_NAME)
            # self.write_parameters(piv_param_grp)
            if recording_dtime is not None:
                ds_rec_dtime = create_recording_datetime_dataset(main, recording_dtime, name='time')

            for i, cname in enumerate(('x', 'y', 'ix', 'iy')):
                ds = main.create_dataset(
                    name=cname,
                    shape=piv_data.data[cname].shape,
                    maxshape=piv_data.data[cname].shape,
                    chunks=piv_data.data[cname].shape,
                    data=piv_data.data[cname], dtype=piv_data.data[cname].dtype,
                    compression=get_config('compression'),
                    compression_opts=get_config('compression_opts'))
                for k, v in piv_data.data_attrs[cname].items():
                    ds.attrs[k] = v
                ds.make_scale(DEFAULT_DATASET_LONG_NAMES[cname])

            for i, cname in enumerate(('z', 'reltime')):
                ds = main.create_dataset(cname, data=piv_data.data[cname])
                for k, v in piv_data.data_attrs[cname].items():
                    ds.attrs[k] = v
                ds.make_scale(DEFAULT_DATASET_LONG_NAMES[cname])

            # Data Arrays
            _shape = (ny, nx)
            for k, v in piv_data.data.items():
                if k not in DIM_NAMES:
                    ds = main.create_dataset(
                        name=k,
                        shape=_shape,
                        dtype=v.dtype,
                        maxshape=_shape,
                        chunks=_shape,
                        compression=get_config('compression'),
                        compression_opts=get_config('compression_opts'))
                    ds[:] = v
                    for ic, c in enumerate((('y', 'iy'), ('x', 'ix'))):
                        ds.dims[ic].attach_scale(main[c[0]])
                        ds.dims[ic].attach_scale(main[c[1]])
                    if recording_dtime is None:
                        ds.attrs['COORDINATES'] = ['reltime', 'z']
                    else:
                        ds.attrs['COORDINATES'] = ['reltime', 'z', ds_rec_dtime.name]

                    if k in piv_data.data_attrs:
                        for attr_key, attr_val in piv_data.data_attrs[k].items():
                            if attr_key not in IGNORE_ATTRS:
                                ds.attrs[attr_key] = attr_val
            # # pivflags explanation:
            # unique_flags = np.unique(main['piv_flags'][:])
            # main['piv_flags'].attrs['flag_meaning'] = json.dumps(
            #     {str(PIVviewFlag(u)): int(u) for u in unique_flags})
        return hdf_filename


class PIVViewStereoNcFile(PIVViewNcFile):
    """Interface class to a PIVview netCDF4 file for stereo recordings. Stereo
    recordings uses different configuration files (.cfg)"""
    __suffix__ = '.nc'
    __parameter_cls__ = pivview_parameter.PivViewConfigFile



import datetime
import pathlib
import shutil
import unittest

from piv2hdf import pivview, tutorial
from piv2hdf.interface import PIVPlane, PIVMultiPlane

__this_dir__ = pathlib.Path(__file__).parent

from piv2hdf.pivview.user_operations import add_standard_name_operation

CREATOR = {
    'name': 'Matthias Probst',
    'id': 'https://orcid.org/0000-0001-8729-0482',
    'role': 'Researcher'}


class TestUser(unittest.TestCase):

    def test_tmpdir_change(self):
        from piv2hdf import user
        default_user_tmpdir = user.TEMPORARY_USER_DIRECTORY
        plane_dirs = tutorial.PIVview.get_multiplane_directories()[0:2]

        # init from folder:
        plane_objs = [PIVPlane.from_folder(
            d,
            time_info=(datetime.datetime.now(), 5),
            pivfile=pivview.PIVViewNcFile,
            user_defined_hdf5_operations=add_standard_name_operation) for d in
            plane_dirs]
        mplane = PIVMultiPlane(plane_objs)
        hdf_filename = mplane.to_hdf(piv_attributes=dict(piv_medium='air', creator=CREATOR))
        self.assertTrue(default_user_tmpdir in hdf_filename.parents)

        new_tmp = __this_dir__ / 'tmp'
        user.TEMPORARY_USER_DIRECTORY = new_tmp
        plane_dirs = tutorial.PIVview.get_multiplane_directories()[0:2]

        # init from folder:
        plane_objs = [PIVPlane.from_folder(d,
                                           time_info=(datetime.datetime.now(), 5),
                                           pivfile=pivview.PIVViewNcFile,
                                           user_defined_hdf5_operations=add_standard_name_operation) for d in
                      plane_dirs]
        mplane = PIVMultiPlane(plane_objs)
        hdf_filename = mplane.to_hdf(piv_attributes=dict(piv_medium='air',
                                                         creator=CREATOR))
        self.assertFalse(default_user_tmpdir in hdf_filename.parents)
        self.assertTrue(new_tmp in hdf_filename.parents)

    def tearDown(self):
        new_tmp = __this_dir__ / 'tmp'
        shutil.rmtree(new_tmp, ignore_errors=True)



import lvpyio as lv
import numpy as np

from piv2hdf.interface import PIVParameterInterface


class DavisParameterFile(PIVParameterInterface):
    __suffix__ = '.vc7'

    def __init__(self, filename):
        super().__init__(filename)
        self._read_file()

    def _read_file(self):
        self._set = lv.read_buffer(str(self.filename.resolve()))
        self._attributes = lv.read_buffer(str(self.filename.resolve()))[0].attributes
        self._piv_param_datasets["x_final_iw_size"]["data"] = self._attributes["InterrogationWindowSize"]
        self._piv_param_datasets["y_final_iw_size"]["data"] = self._attributes["InterrogationWindowSize"]
        for k, v in self._attributes.items():
            if isinstance(v, str):
                if "s" in v:
                    value, unit = v.split(" ", 1)
                    try:
                        self._piv_param_datasets[k] = dict(data=float(value), unit=unit)
                    except ValueError:
                        self._piv_param_datasets[k] = dict(data=value, unit=unit)
                else:
                    try:
                        value = float(v)
                    except ValueError:
                        value = v
                    self._piv_param_datasets[k] = dict(data=value)
            else:
                if isinstance(v, np.ndarray):
                    self._piv_param_datasets[k] = dict(data=v[0][0])
                else:
                    self._piv_param_datasets[k] = dict(data=v)

    @classmethod
    def from_dir(cls, directory_path):
        raise NotImplementedError("DavisParameterFile should be instantiated with a filename")



import datetime
import json
import pathlib
import unittest

import h5py
import h5rdmtoolbox as h5tbx
import numpy as np

from piv2hdf import pivview, openpiv, tutorial
from piv2hdf import set_loglevel
from piv2hdf.interface import PIVPlane, PIVMultiPlane
from piv2hdf.openpiv.user_operations import add_standard_name_operation as opernpiv_add_standard_name_operation
from piv2hdf.pivview.user_operations import add_standard_name_operation

NOW = datetime.datetime.now()

__this_dir__ = pathlib.Path(__file__).parent


def read_meta():
    with open(__this_dir__ / "resources/meta.json") as f:
        meta = json.load(f)
    return meta


class TestMPlane(unittest.TestCase):

    def setUp(self):
        set_loglevel("DEBUG")
        self.meta = read_meta()

    def test_multiplane_different_pivfile(self):
        plane_dirs_pivview = tutorial.PIVview.get_multiplane_directories()[0:2]
        plane_dirs_openpiv = tutorial.OpenPIV.get_multiplane_directories()[0:2]

        # pass wrong folders which will raise an error in the process:
        with self.assertRaises(ValueError):
            _ = PIVPlane.from_folder(plane_dirs_pivview[0], time_info=(NOW, 5), pivfile=openpiv.OpenPIVFile,
                                     user_defined_hdf5_operations=opernpiv_add_standard_name_operation)
        with self.assertRaises(ValueError):
            _ = PIVPlane.from_folder(plane_dirs_openpiv[1], time_info=(NOW, 5), pivfile=pivview.PIVViewNcFile,
                                     user_defined_hdf5_operations=add_standard_name_operation)

        plane0 = PIVPlane.from_folder(plane_dirs_openpiv[0], time_info=(NOW, 5), pivfile=openpiv.OpenPIVFile,
                                      user_defined_hdf5_operations=opernpiv_add_standard_name_operation)
        plane1 = PIVPlane.from_folder(plane_dirs_pivview[1], time_info=(NOW, 5), pivfile=pivview.PIVViewNcFile,
                                      user_defined_hdf5_operations=add_standard_name_operation)
        with self.assertWarns(UserWarning):
            _ = PIVMultiPlane([plane0, plane1])

    def test_multi_piv_identical_nt_pivview(self):
        plane_dirs = tutorial.PIVview.get_multiplane_directories()[0:2]

        # init from folder:
        dt_start1 = datetime.datetime(2018, 1, 1)
        dt_start2 = datetime.datetime(2018, 1, 1) + datetime.timedelta(minutes=10)
        dt_start3 = datetime.datetime(2018, 1, 1) + datetime.timedelta(minutes=20)

        time_vector2 = [dt_start2 + datetime.timedelta(seconds=i / 10) for i in range(100)]
        time_infos = [(dt_start1, 10), time_vector2, (dt_start3, 10)]

        plane_objs = [PIVPlane.from_folder(d,
                                           pivfile=pivview.PIVViewNcFile,
                                           user_defined_hdf5_operations=add_standard_name_operation,
                                           time_info=time_info) for d, time_info in
                      zip(plane_dirs, time_infos)]
        mplane = PIVMultiPlane(plane_objs)
        hdf_filename = mplane.to_hdf(
            piv_attributes=dict(piv_medium=self.meta["PIV_MEDIUM"], creator=self.meta["CREATOR"]))
        with h5tbx.File(hdf_filename) as h5:

            nt = h5['u'].shape[1]
            for k in ('reltime', 'image_index'):
                self.assertEqual(nt, h5[k].shape[0])

            self.assertEqual(nt, h5['time'].shape[1])

            self.assertEqual(np.datetime64(dt_start1), h5.time[0, 0].values)
            self.assertEqual(np.datetime64(dt_start2), h5.time[1, 0].values)
            for i in range(1, nt):
                self.assertEqual(np.datetime64(dt_start1 + datetime.timedelta(seconds=i / 10)),
                                 h5.time[0, i].values)
                self.assertEqual(np.datetime64(dt_start2 + datetime.timedelta(seconds=i / 10)),
                                 h5.time[1, i].values)

            for k, v in h5.items():
                if isinstance(v, h5py.Dataset) and v.ndim == 4:
                    self.assertEqual('/z', v.dims[0][0].name)
                    # self.assertEqual('/iz', v.dims[0][1].name)
                    self.assertEqual('/reltime', v.dims[1][0].name)
                    # self.assertEqual('/time', v.dims[1][1].name)
                    self.assertEqual('/image_index', v.dims[1][1].name)
                    self.assertEqual('/y', v.dims[2][0].name)
                    self.assertEqual('/iy', v.dims[2][1].name)
                    self.assertEqual('/x', v.dims[3][0].name)
                    self.assertEqual('/ix', v.dims[3][1].name)

            self.assertIn('software', h5.attrs)
            self.assertEqual(h5['x'].size, 31)
            self.assertEqual(h5['y'].size, 15)
            self.assertEqual(h5['z'].size, 2)
            self.assertEqual(list(h5['z'][()]), [-5., 0.])
            self.assertEqual(h5['u'].shape, (2, 3, 15, 31))
            self.assertEqual(h5['v'].shape, (2, 3, 15, 31))
            self.assertIsNotNone(h5.find_one({'standard_name': {'$regex': 'piv_flag'}}))
            self.assertEqual('uint8', h5.find_one({'standard_name': {'$regex': 'piv_flag'}}).dtype)
            for k, v in h5.items():
                if isinstance(v, h5py.Dataset) and v.ndim == 4:
                    if v.dims[0][0] != h5['z']:
                        raise ValueError(f"Expected {v.dims[0][0]} == {h5['z']}")
                    if v.dims[1][0] != h5['reltime']:
                        raise ValueError(f"Expected {v.dims[1][0]} == {h5['reltime']}")
                    if v.dims[2][0] != h5['y']:
                        raise ValueError(f"Expected {v.dims[2][0]} == {h5['y']}")
                    if v.dims[3][0] != h5['x']:
                        raise ValueError(f"Expected {v.dims[3][0]} == {h5['x']}")

    def test_multi_piv_similar_nt_pivview(self):
        plane_dirs = tutorial.PIVview.get_multiplane_directories()[0:2]

        # init from folder:
        dt_start1 = datetime.datetime(2018, 1, 1)
        dt_start2 = datetime.datetime(2018, 1, 1) + datetime.timedelta(minutes=10)
        dt_start3 = datetime.datetime(2018, 1, 1) + datetime.timedelta(minutes=20)
        time_infos = [(dt_start1, 10), (dt_start2, 10), (dt_start3, 10)]

        plane_objs = [PIVPlane.from_folder(d,
                                           pivfile=pivview.PIVViewNcFile,
                                           user_defined_hdf5_operations=add_standard_name_operation,
                                           time_info=time_info) for d, time_info in
                      zip(plane_dirs, time_infos)]
        mplane = PIVMultiPlane(plane_objs)
        hdf_filename = mplane.to_hdf(
            piv_attributes=dict(piv_medium=self.meta["PIV_MEDIUM"], creator=self.meta["CREATOR"]),
            atol=1e-3, rtol=1e-3)
        self.assertEqual(hdf_filename, mplane.hdf_filename)
        with h5tbx.File(hdf_filename) as h5:
            self.assertIn('software', h5.attrs)
            # self.assertTrue(
            #     np.array_equal(h5['reltime'][()], np.asarray(rtfs).mean(axis=0).astype(h5['reltime'].dtype)))

            for k, v in h5.items():
                if isinstance(v, h5py.Dataset) and v.ndim == 4:
                    self.assertEqual('/z', v.dims[0][0].name)
                    # self.assertEqual('/iz', v.dims[0][1].name)
                    self.assertEqual('/reltime', v.dims[1][0].name)
                    # self.assertEqual('/time', v.dims[1][1].name)
                    self.assertEqual('/image_index', v.dims[1][1].name)
                    self.assertEqual(1, v.dims[1][1][0])
                    self.assertEqual('/y', v.dims[2][0].name)
                    self.assertEqual('/iy', v.dims[2][1].name)
                    self.assertEqual('/x', v.dims[3][0].name)
                    self.assertEqual('/ix', v.dims[3][1].name)

            self.assertEqual(h5['x'].size, 31)
            self.assertEqual(h5['y'].size, 15)
            self.assertEqual(h5['z'].size, 2)
            self.assertEqual(list(h5['z'][()]), [-5., 0.])
            self.assertEqual(h5['u'].shape, (2, 3, 15, 31))
            self.assertEqual(h5['v'].shape, (2, 3, 15, 31))
            self.assertIsNotNone(h5.find_one({'standard_name': {'$regex': 'piv_flag'}}))
            self.assertEqual('uint8', h5.find_one({'standard_name': {'$regex': 'piv_flag'}}).dtype)
            for k, v in h5.items():
                if isinstance(v, h5py.Dataset) and v.ndim == 4:
                    assert v.dims[0][0] == h5['z']
                    assert v.dims[1][0] == h5['reltime']
                    assert v.dims[2][0] == h5['y']
                    assert v.dims[3][0] == h5['x']

    def test_multi_piv_similar_time_but_different_nt_pivview(self):
        plane_dirs = tutorial.PIVview.get_multiplane_directories()[0:3]

        # init from folder:
        dt_start1 = datetime.datetime(2018, 1, 1)
        dt_start2 = datetime.datetime(2018, 1, 1, 1, 0, 0)
        dt_start3 = datetime.datetime(2018, 1, 1, 2, 0, 0)
        time_infos = [(dt_start1,
                       dt_start1 + datetime.timedelta(seconds=0.1),
                       dt_start1 + datetime.timedelta(seconds=0.3)),
                      (dt_start2,
                       dt_start2 + datetime.timedelta(seconds=0.1),
                       dt_start2 + datetime.timedelta(seconds=0.3)),
                      (dt_start3,
                       dt_start3 + datetime.timedelta(seconds=0.1)), ]

        plane_objs = [PIVPlane.from_folder(d,
                                           pivfile=pivview.PIVViewNcFile,
                                           time_info=time_info,
                                           user_defined_hdf5_operations=add_standard_name_operation,
                                           ) for d, time_info in
                      zip(plane_dirs, time_infos)]
        mplane = PIVMultiPlane(plane_objs)
        hdf_filename = mplane.to_hdf(
            piv_attributes=dict(piv_medium=self.meta["PIV_MEDIUM"], creator=self.meta["CREATOR"]),
            atol=1e-1, rtol=1e-1)
        with h5tbx.File(hdf_filename) as h5:
            self.assertIn('software', h5.attrs)
            # self.assertTrue(np.array_equal(h5['reltime'][()],
            #                                np.asarray([t[:2] for t in rtfs]).mean(axis=0).astype(h5['reltime'].dtype)))

            for k, v in h5.items():
                if isinstance(v, h5py.Dataset) and v.ndim == 4:
                    self.assertEqual('/z', v.dims[0][0].name)
                    # self.assertEqual('/iz', v.dims[0][1].name)
                    self.assertEqual('/reltime', v.dims[1][0].name)
                    # self.assertEqual('/time', v.dims[1][1].name)
                    self.assertEqual('/image_index', v.dims[1][1].name)
                    self.assertEqual(1, v.dims[1][1][0])
                    self.assertEqual('/y', v.dims[2][0].name)
                    self.assertEqual('/iy', v.dims[2][1].name)
                    self.assertEqual('/x', v.dims[3][0].name)
                    self.assertEqual('/ix', v.dims[3][1].name)

            self.assertEqual(h5['x'].size, 31)
            self.assertEqual(h5['y'].size, 15)
            self.assertEqual(h5['z'].size, 3)
            self.assertEqual(list(h5['z'][()]), [-5., 0., 10.0])
            self.assertEqual(h5['u'].shape, (3, 2, 15, 31))
            self.assertEqual(h5['v'].shape, (3, 2, 15, 31))
            self.assertIsNotNone(h5.find_one({'standard_name': {'$regex': 'piv_flag'}}))
            self.assertEqual('uint8', h5.find_one({'standard_name': {'$regex': 'piv_flag'}}).dtype)
            for k, v in h5.items():
                if isinstance(v, h5py.Dataset) and v.ndim == 4:
                    assert v.dims[0][0] == h5['z']
                    assert v.dims[1][0] == h5['reltime']
                    assert v.dims[2][0] == h5['y']
                    assert v.dims[3][0] == h5['x']

    def test_unequal_time_vectors_pivview(self):
        plane_dirs = tutorial.PIVview.get_multiplane_directories()[0:3]

        dt_start1 = datetime.datetime(2018, 1, 1)
        dt_start2 = datetime.datetime(2018, 1, 1) + datetime.timedelta(minutes=5)
        dt_start3 = datetime.datetime(2018, 1, 1) + datetime.timedelta(minutes=20)
        rtfs = [(dt_start1, 10), (dt_start2, 8), (dt_start3, 10)]  # different freq!
        plane_objs = [PIVPlane.from_folder(d,
                                           pivfile=pivview.PIVViewNcFile,
                                           user_defined_hdf5_operations=add_standard_name_operation,
                                           time_info=time_info) for d, time_info in
                      zip(plane_dirs, rtfs)]
        mplane = PIVMultiPlane(plane_objs)

        with self.assertRaises(ValueError):
            # time vectors are not close enough
            mplane.to_hdf(piv_attributes=dict(piv_medium=self.meta["PIV_MEDIUM"],
                                              creator=self.meta["CREATOR"]),
                          atol=1e-3, rtol=1e-3)

    def test_multi_piv_equal_nt_openpiv(self):
        plane_dirs = tutorial.OpenPIV.get_multiplane_directories()[0:2]
        plane_objs = [PIVPlane.from_folder(d,
                                           time_info=(NOW, 5),
                                           pivfile=openpiv.OpenPIVFile,
                                           user_defined_hdf5_operations=opernpiv_add_standard_name_operation) for d in
                      plane_dirs]
        mplane = PIVMultiPlane(plane_objs)
        hdf_filename = mplane.to_hdf(piv_attributes=dict(piv_medium=self.meta["PIV_MEDIUM"],
                                                         creator=self.meta["CREATOR"]),
                                     z=[-5., 0.])
        with h5tbx.File(hdf_filename) as h5:

            for k, v in h5.items():
                if isinstance(v, h5py.Dataset) and v.ndim == 4:
                    self.assertEqual('/z', v.dims[0][0].name)
                    # self.assertEqual('/iz', v.dims[0][1].name)
                    self.assertEqual('/reltime', v.dims[1][0].name)
                    # self.assertEqual('/time', v.dims[1][1].name)
                    self.assertEqual('/image_index', v.dims[1][1].name)
                    self.assertEqual(1, v.dims[1][1][0])
                    self.assertEqual('/y', v.dims[2][0].name)
                    self.assertEqual('/iy', v.dims[2][1].name)
                    self.assertEqual('/x', v.dims[3][0].name)
                    self.assertEqual('/ix', v.dims[3][1].name)
            self.assertIn('software', h5.attrs)
            self.assertEqual(h5['x'].size, 63)
            self.assertEqual(h5['y'].size, 59)
            self.assertEqual(h5['z'].size, 2)
            self.assertEqual(list(h5['z'][()]), [-5., 0.])
            self.assertEqual(h5['u'].shape, (2, 3, 59, 63))
            self.assertEqual(h5['v'].shape, (2, 3, 59, 63))
            self.assertIsNotNone(h5.find_one({'standard_name': {'$regex': 'piv_flag'}}))
            self.assertEqual('uint8', h5.find_one({'standard_name': {'$regex': 'piv_flag'}}).dtype)

    def test_multi_piv_unequal_nt_pivview(self):
        from piv2hdf import PIVMultiPlane
        from piv2hdf.pivview import PIVViewNcFile
        from piv2hdf import tutorial

        multi_plane_dirs = tutorial.PIVview.get_multiplane_directories()
        mplane = PIVMultiPlane.from_folders(plane_directories=multi_plane_dirs,
                                            time_infos=[(NOW, 5), (NOW, 10), (NOW, 3)],
                                            pivfile=PIVViewNcFile,
                                            user_defined_hdf5_operations=add_standard_name_operation)
        with self.assertRaises(ValueError):
            mplane.to_hdf(
                piv_attributes=dict(piv_medium=self.meta["PIV_MEDIUM"], creator=self.meta["CREATOR"]),
            )

        multi_plane_dirs = tutorial.PIVview.get_multiplane_directories()
        mplane = PIVMultiPlane.from_folders(plane_directories=multi_plane_dirs,
                                            time_infos=[(NOW, 5), (NOW, 5.001), (NOW, 4.9999)],
                                            pivfile=PIVViewNcFile,
                                            user_defined_hdf5_operations=add_standard_name_operation)
        hdf_filename = mplane.to_hdf(
            piv_attributes=dict(piv_medium=self.meta["PIV_MEDIUM"], creator=self.meta["CREATOR"]),
            atol=0.1, rtol=0.1
        )
        with h5tbx.File(hdf_filename) as h5:
            self.assertIsNotNone(h5.find_one({'standard_name': {'$regex': 'piv_flag'}}))
            self.assertEqual('uint8', h5.find_one({'standard_name': {'$regex': 'piv_flag'}}).dtype)
            for plane_name in ('plane0', 'plane1', 'plane2'):
                for k, v in h5[plane_name].items():
                    if isinstance(v, h5py.Dataset) and v.ndim == 3:
                        assert v.dims[0][0] == v.parent['reltime']
                        assert v.dims[1][0] == h5['y']
                        assert v.dims[2][0] == h5['x']

            for k, v in h5.items():
                if isinstance(v, h5py.Dataset) and v.ndim == 4:
                    self.assertEqual('/z', v.dims[0][0].name)
                    # self.assertEqual('/iz', v.dims[0][1].name)
                    self.assertEqual('/reltime', v.dims[1][0].name)
                    # self.assertEqual('/time', v.dims[1][1].name)
                    self.assertEqual('/image_index', v.dims[1][1].name)
                    self.assertEqual(1, v.dims[1][1][0])
                    self.assertEqual('/y', v.dims[2][0].name)
                    self.assertEqual('/iy', v.dims[2][1].name)
                    self.assertEqual('/x', v.dims[3][0].name)
                    self.assertEqual('/ix', v.dims[3][1].name)



import abc
import datetime
import json
import os
import pathlib
import warnings
from functools import wraps
from typing import List, Union, Dict, Tuple, Callable, Type, Protocol, Optional, runtime_checkable

import h5py
import h5rdmtoolbox as h5tbx
import numpy as np
from tqdm import tqdm

from piv2hdf.config import get_config, get_pivattrs
from . import postproc
from ._logger import logger
from .time import TimeVectorWarning
from .time import create_recording_datetime_dataset
from .utils import generate_temporary_filename, get_uint_type, parse_z
from dataclasses import dataclass

PIV_PARAMETER_GRP_NAME = 'piv_parameters'
PIV_PARAMETER_ATTRS_NAME = 'piv_parameter_dict'

DEFAULT_MPLANE_TITLE = 'piv snapshot data'
DEFAULT_PLANE_TITLE = 'piv plane data'

IMAGE_INDEX = 'image_index'  # name of the image index coordinate


class LayoutValidationWarning(Warning):
    """Warning for layout validation"""
    pass


def useroperation(func: Callable) -> Callable:
    """decorates to_hdf method of Snapshot, Plane and Multi-Plane, that runs user code to
    update metadata after the data has been written to the HDF5 file and before layout validation"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        """call __to_hdf__ and then perform the layout validation. Raise an error on failure"""
        hdf_filename = func(*args, **kwargs)
        user_defined_hdf5_operations = args[0].__get_user_defined_hdf5_operations__()
        if user_defined_hdf5_operations:
            with h5tbx.File(hdf_filename, 'r+') as h5:
                for pc in user_defined_hdf5_operations:
                    logger.debug(f'Calling user defined hdf5 operation "{pc.__class__.__name__}"...')
                    pc(h5)

        for pp_func_name in get_config()['postproc']:
            if callable(pp_func_name):
                pp_func = pp_func_name
            elif isinstance(pp_func_name, str):
                if pp_func_name not in dir(postproc):
                    raise NotImplementedError(f'Postprocessing function {pp_func_name} not implemented.')
                pp_func = getattr(postproc, pp_func_name)
            else:
                raise TypeError(f'Expected a string or callable for key "postproc", got {type(pp_func_name)}')
            with h5tbx.File(hdf_filename, 'r+') as h5:
                pp_func(h5)
        return hdf_filename

    return wrapper


def layoutvalidation(func: Callable) -> Callable:
    """decorates __to_hdf__ method of Snapshot, Plane and Multi-Plane classes to validate the created
    HDF5 content"""

    @wraps(func)
    def validation_wrapper(*args, **kwargs):
        """call __to_hdf__ and then perform the layout validation. Raise an error on failure"""
        hdf_filename = func(*args, **kwargs)

        # checking standard attributes:

        with h5tbx.use(get_config('convention')) as cv:
            logger.debug(
                f'Checking the PIV attributes based on convention currently set ({cv.name} @ {cv.filename})...'
            )
            with h5tbx.File(hdf_filename, 'r+') as h5:

                for sa_name, std_attr in h5.standard_attributes.items():
                    if sa_name not in h5.attrs.raw:
                        if std_attr.default_value == std_attr.__class__.NONE:
                            logger.debug(f'Attribute {sa_name} not set but is also only optional.')
                            continue
                        else:
                            logger.error(f'Attribute not set: {sa_name} but is required.')
                            continue

                    std_attr.validate(h5.attrs.raw[sa_name], parent=h5)

                if 'standard_name_table' in h5.standard_attributes:
                    snt = h5.standard_name_table
                    snt.check_hdf_group(h5)

        return hdf_filename

    return validation_wrapper


def scan_for_timeseries_files(directory: pathlib.Path, suffix: str, prefix_pattern: None) -> List[pathlib.Path]:
    """
    Scans for files in `directory` according to the `suffix` provided. With `prefix_pattern` the pattern can be
    specified. `prefix_pattern` and `suffix` are concatenated to form the final pattern, which is passed to
    `pathlib.Path.glob()`. The result is a list of files sorted by name.
    """
    if prefix_pattern is None:
        prefix_pattern = '*'
    logger.debug(f'Scanning for files with pattern {prefix_pattern}{suffix} in {directory}')
    list_of_files = sorted(directory.glob(f'{prefix_pattern}{suffix}'))
    return list_of_files


def copy_piv_parameter_group(src: h5py.Group, trg: h5py.Group) -> None:
    """copies the piv parameters to the new group"""

    def _to_grp(_src, _trg):
        for ak, av in _src.attrs.items():
            _trg.attrs[ak] = av
        for k, v in _src.items():
            if isinstance(v, h5py.Group):
                obj = _trg.create_group(k)
                _to_grp(v, obj)
            else:
                obj = _trg.create_dataset(k, data=v[()])
            for ak, av in v.attrs.items():
                obj.attrs[ak] = av

    _to_grp(src, trg)


class PIVParameterInterface:
    """Abstract PIV Parameter Interface"""
    __slots__ = ('param_dict', '_piv_param_datasets', 'filename')
    __suffix__ = None

    def __init__(self, filename):
        if filename is not None:
            self.filename = pathlib.Path(filename)
            if not self.filename.exists():
                raise FileNotFoundError(f'File {filename} does not exist.')
        else:
            self.filename = filename
        self.param_dict = {}
        self._piv_param_datasets = dict(x_final_iw_size=dict(data=None,
                                                             units='pixel',
                                                             # standard_name='x_final_interrogation_window_size',
                                                             long_name='Final interrogation window size in x-direction'),
                                        y_final_iw_size=dict(data=None,
                                                             units='pixel',
                                                             # standard_name='y_final_interrogation_window_size',
                                                             long_name='Final interrogation window size in y-direction'),
                                        x_final_iw_overlap_size=dict(data=None,
                                                                     units='pixel',
                                                                     # standard_name='x_final_interrogation_window_overlap_size',
                                                                     long_name='Final overlap of interrogation windows in x-direction'),
                                        y_final_iw_overlap_size=dict(data=None,
                                                                     units='pixel',
                                                                     # standard_name='y_final_interrogation_window_overlap_size',
                                                                     long_name='Final overlap of interrogation windows in y-direction'),
                                        laser_pulse_delay=dict(data=None,
                                                               units='s',
                                                               # standard_name='laser_pulse_delay',
                                                               long_name='Pulse delay'),
                                        piv_scaling_factor=dict(data=None,
                                                                units='pixel/m',
                                                                # standard_name='piv_scaling_factor',
                                                                )
                                        )

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.filename}": {self.param_dict})'

    @property
    def piv_param_datasets(self):
        return self._piv_param_datasets

    @property
    def suffix(self):
        """Return the suffix of the PIV Parameter File"""
        return self.__suffix__

    @classmethod
    def from_dir(cls, directory_path):
        """reads parameter from `directory_path` and returns instance of this class

        Parameters
        ----------
        directory_path: pathlib.Path
            Path to (plane) directory containing the .nc-files

        Returns
        -------
        Instance of this class
        """
        pars = list(directory_path.glob(f'*{cls.__suffix__}'))
        if len(pars) > 1:
            raise FileExistsError(f'More than one parameter file with suffix "{cls.__suffix__}" '
                                  f'found here: "{directory_path}".')
        elif len(pars) == 0:
            raise FileExistsError(f'Cannot find the parameter file with suffix "{cls.__suffix__}" '
                                  f'here: "{directory_path}".')
        return cls(pars[0])

    @abc.abstractmethod
    def save(self, filename: pathlib.Path):
        """Save to original file format"""

    def to_dict(self) -> Dict:
        """Convert to a dictionary"""
        return self.param_dict

    def to_hdf(self, grp: h5py.Group) -> h5py.Group:
        """Write parameter content to HDF group"""
        grp.attrs[PIV_PARAMETER_ATTRS_NAME] = json.dumps(self.param_dict)
        for k, v in self.piv_param_datasets.items():
            if v['data']:
                ds = grp.create_dataset(name=k, data=v['data'])
                for ak, av in v.items():
                    if ak != 'data':
                        ds.attrs[ak] = av
        return grp

@dataclass
class PIVData:
    data: Dict
    data_attrs: Dict
    root_attrs: Dict

@runtime_checkable
class UserDefinedHDF5Operation(Protocol):
    """Protocol class enabling users to inject file manipulation after
    HDF5 file conversion.
    The only method which needs to be provided by the class is __call__.
    """

    def __call__(self, h5: h5tbx.File):
        """This method manipulates the HDF5 file"""


class PIVFile(abc.ABC):
    """Abstract Interface class for PIV. This class must be inherited for each PIV file type.
    Required methods are:
        - read(recording_dtime: float, build_coord_datasets: bool = True)
        - to_hdf(hdf_filename: pathlib.Path, recording_dtime: float)
    """
    suffix: str = None
    __parameter_cls__ = PIVParameterInterface

    def __init__(self,
                 filename: pathlib.Path,
                 parameter: Union[None, PIVParameterInterface, pathlib.Path] = None,
                 user_defined_hdf5_operations: Optional[
                     Union[UserDefinedHDF5Operation, List[UserDefinedHDF5Operation]]] = None,
                 **kwargs):
        filename = pathlib.Path(filename)
        if not filename.is_file():
            raise TypeError(f'Snapshot file path is not a file: {filename}.')
        if self.suffix is not None:
            if self.suffix != '':
                if filename.suffix != self.suffix:
                    raise NameError(f'Expecting suffix {self.suffix}, not {filename.suffix}')
        self.filename = filename
        if isinstance(parameter, PIVParameterInterface):
            self._parameter = parameter
        elif parameter is None:
            logger.debug(
                f'Auto detecting parameter file {self.__parameter_cls__.__name__} from directory {filename.parent}')
            self._parameter = self.__parameter_cls__.from_dir(filename.parent)
        else:
            logger.debug(f'Initializing parameter file from filename using class {self.__parameter_cls__.__name__}')
            self._parameter = self.__parameter_cls__(parameter)

        if user_defined_hdf5_operations is None:
            user_defined_hdf5_operations = []
        if not isinstance(user_defined_hdf5_operations, (tuple, list)):
            user_defined_hdf5_operations = [user_defined_hdf5_operations, ]
        for udo in user_defined_hdf5_operations:
            if not isinstance(udo, UserDefinedHDF5Operation):
                raise TypeError(f'Expecting type {UserDefinedHDF5Operation.__name__}, not {type(udo)}')
        self.user_defined_hdf5_operations = user_defined_hdf5_operations
        #
        # if callable(post_func):
        #     self.post_func = [post_func, ]
        # elif isinstance(post_func, (list, tuple)):
        #     self.post_func = post_func
        # elif post_func is None:
        #     self.post_func = []
        # else:
        #     raise TypeError('post_func must be callable or list/tuple of callables or None')

    @property
    def parameter(self):
        """Return the parameter class"""
        return self._parameter

    def __repr__(self):
        return f'<{self.__class__.__name__} ({self.filename})>'

    @abc.abstractmethod
    def read(
            self,
            relative_time: float,
            **kwargs
    ) -> PIVData:
        """Read data from file.
        Except data, root_attr, variable_attr"""

    def write_parameters(self, param_grp: h5py.Group):
        """Write piv parameters to an opened and existing param_grp"""
        return self._parameter.to_hdf(param_grp)

    @abc.abstractmethod
    def to_hdf(self,
               hdf_filename: pathlib.Path,
               relative_time: float,
               recording_dtime: Union[datetime.datetime, List[datetime.datetime]],
               z: Union[str, float, None] = None,
               **kwargs) -> pathlib.Path:
        """converts the snapshot into an HDF file"""


class PIVConverter(abc.ABC):
    """Abstract converter class"""

    def __init__(self):
        self.hdf_filename = None

    def __repr__(self):
        return f'<{self.__class__.__name__}>'

    @abc.abstractmethod
    def __get_user_defined_hdf5_operations__(self) -> List[UserDefinedHDF5Operation]:
        """return the user defined hdf5 operation class"""

    @abc.abstractmethod
    def __to_hdf__(self, hdf_filename: pathlib.Path, **kwargs) -> pathlib.Path:
        """conversion method"""

    @layoutvalidation
    @useroperation
    def to_hdf(self,
               piv_attributes: Dict = None,  # e.g. piv_medium, contact
               hdf_filename: pathlib.Path = None,
               z: float = None,
               **kwargs) -> pathlib.Path:
        """conversion method

        Parameters
        ----------
        piv_attributes: Dict
            Attributes to be added to the root group of the HDF file
        hdf_filename: pathlib.Path
            Path to the HDF file to be written. None will generate an appropriate filename.
            If the path is a directory, the filename will be generated from the directory name.
        z: float
            z-coordinate of the plane. If None, the z-coordinate is read from the parameter file
        kwargs: Dict
            Additional keyword arguments for the conversion method
        """
        if hdf_filename is not None:
            _hdf_filename = pathlib.Path(hdf_filename)
            if _hdf_filename.is_dir():
                hdf_filename = _hdf_filename / f'{_hdf_filename.stem}.hdf'
        hdf_filename = self.__to_hdf__(hdf_filename=hdf_filename,
                                       z=z,
                                       **kwargs)
        self.hdf_filename = hdf_filename

        _piv_attrs = get_pivattrs()

        if piv_attributes:
            _piv_attrs.update(piv_attributes)

        with h5tbx.File(hdf_filename, 'r+') as h5:
            for k, v in _piv_attrs.items():
                h5.attrs[k] = v

        return hdf_filename


class PIVSnapshot(PIVConverter):
    """Interface class"""

    def __init__(self,
                 piv_file: PIVFile,
                 recording_dtime: Union[datetime.datetime, None]):
        super().__init__()
        if not isinstance(piv_file, PIVFile):
            raise TypeError(f'Expecting type {PIVFile.__class__} fro parameter piv_file, not type {type(piv_file)}')

        # process recording_dtime:
        if recording_dtime is None:
            logger.debug('No datetime provided!')
            self.recording_dtime = None
        else:
            if not isinstance(recording_dtime, datetime.datetime):
                raise TypeError('Expecting type int, float or datetime for recording time, '
                                f'not type {type(recording_dtime)}')

            self.recording_dtime = recording_dtime  # .isoformat()
        self.relative_time = 0
        self.piv_file = piv_file

    def __repr__(self):
        return f'<{self.__class__.__name__} of {self.piv_file.__repr__()} >'

    def __get_user_defined_hdf5_operations__(self) -> List[UserDefinedHDF5Operation]:
        return self.piv_file.user_defined_hdf5_operations

    def __to_hdf__(self,
                   hdf_filename: pathlib.Path = None,
                   z: Union[str, float, None] = None,
                   **kwargs) -> pathlib.Path:
        """converts the snapshot into an HDF file
        """

        if hdf_filename is None:
            hdf_filename = self.piv_file.filename.parent / f'{self.piv_file.filename.stem}.hdf'
        else:
            hdf_filename = pathlib.Path(hdf_filename)
            hdf_filename.parent.mkdir(parents=True, exist_ok=True)
        hdf_filename = self.piv_file.to_hdf(hdf_filename=hdf_filename,
                                            relative_time=self.relative_time,
                                            recording_dtime=self.recording_dtime,
                                            z=z)
        with h5py.File(hdf_filename, 'r+') as h5:
            self.piv_file.write_parameters(h5.create_group(PIV_PARAMETER_GRP_NAME))
        return hdf_filename


class LayoutValidationError(Exception):
    """Exception raised when Layout validation failed"""


class PIVPlane(PIVConverter):
    """Interface class"""

    __slots__ = 'list_of_piv_files', 'rel_time_vector'
    plane_coord_order = ('reltime', 'y', 'x')

    def __init__(self,
                 list_of_piv_files: List[PIVFile],
                 time_info: Union[Tuple[datetime.datetime, float], List[datetime.datetime]]):
        """

        Parameters
        ----------
        list_of_piv_files: List[PIVFile]
            List of PIVFile objects to be combined
        time_info: Union[Tuple[datetime.datetime, float], List[datetime.datetime]]
            Time information based on which the relative and absolute time vectors are generated.
            The input can be two-fold:
            1. A Tuple of datetime object specifying the start of the PIV recording and a float specifying the
            measurement frequency
            2. A list of datetime objects corresponding to the recording datetime of each image
        """
        super().__init__()
        if not isinstance(list_of_piv_files, (tuple, list)):
            raise TypeError(f'Expecting a list of {PIVFile.__class__} objects but got {type(list_of_piv_files)}')
        if not all(isinstance(piv_file, PIVFile) for piv_file in list_of_piv_files):
            for piv_file in list_of_piv_files:
                if not isinstance(piv_file, PIVFile):
                    raise TypeError(
                        f'Expecting type {PIVFile.__class__} for each entry, but one entry is of type {type(piv_file)}')
        self.list_of_piv_files = list_of_piv_files
        n = len(list_of_piv_files)
        if n == 0:
            raise ValueError('List of piv files is empty!')
        if n == 1:
            raise ValueError('Only one piv file is provided. Use PIVSnapshot instead!')

        # process recording_dtime:
        if not isinstance(time_info, (tuple, list)):
            raise TypeError(f'Expecting type tuple or list for parameter time_info, '
                            f'not type {type(time_info)}')
        if len(time_info) == 2:
            if isinstance(time_info[1], (float, int)):
                # build time vector based on recording start datetime and frequency
                recording_start_dtime, meas_frequency = time_info
                if not isinstance(recording_start_dtime, datetime.datetime):
                    raise TypeError(f'Expecting type datetime for parameter time_info[0], '
                                    f'not type {type(time_info[0])}')
                # build time vector (rel and abs)
                rel_time = np.linspace(0, (n - 1) / meas_frequency, n)
                abs_time = [recording_start_dtime + datetime.timedelta(seconds=float(rel_time[i])) for i in range(n)]
            else:
                if not all(isinstance(r, datetime.datetime) for r in time_info):
                    raise TypeError('Expecting type datetime for all items of parameter time_info[0]')
                abs_time = time_info
                rel_time = [(abs_time[i] - abs_time[0]).total_seconds() for i in range(n)]
        else:
            # expecting a list of datetime objects
            if not all(isinstance(r, datetime.datetime) for r in time_info):
                raise TypeError('Expecting type datetime for all items of parameter time_info[0]')
            if n != len(time_info):
                warnings.warn(f'Length of time_info ({len(time_info)}) is inconsistent with number of piv files ({n}). '
                              f'Assuming that time_info is a correct time vector, thus taking only the first {n} '
                              'entries.', TimeVectorWarning)
            abs_time = time_info[0:n]
            rel_time = [(abs_time[i] - abs_time[0]).total_seconds() for i in range(n)]

        self.rel_time_vector = np.asarray(rel_time)
        self.abs_time_vector = np.asarray(abs_time)
        self.IMAGE_INDEX = np.arange(1, n + 1, 1)

        if len(self.rel_time_vector) != len(list_of_piv_files):
            raise ValueError(f'Length of time vector ({len(self.rel_time_vector)}) is inconsistent with number of piv '
                             f'files: ({len(list_of_piv_files)})!')

    def __get_user_defined_hdf5_operations__(self) -> List[UserDefinedHDF5Operation]:
        return self.list_of_piv_files[0].user_defined_hdf5_operations

    def __getitem__(self, item):
        return self.list_of_piv_files[item]

    def __len__(self):
        return len(self.list_of_piv_files)

    def __repr__(self):
        if len(self.list_of_piv_files) > 0:
            return f'<{self.__class__.__name__} containing {len(self.list_of_piv_files)} ' \
                   f'<{self.list_of_piv_files[0].__class__.__name__}> files>'
        return f'<Empty {self.__class__.__name__}>'

    @staticmethod
    def from_folder(plane_directory: Union[str, pathlib.Path],
                    *,
                    time_info: Union[Tuple[datetime.datetime, float], List[datetime.datetime]],
                    pivfile: Type[PIVFile],
                    user_defined_hdf5_operations: Optional[
                        Union[UserDefinedHDF5Operation, List[UserDefinedHDF5Operation]]
                    ] = None,
                    parameter: Union[PIVParameterInterface, str, pathlib.Path] = None,
                    n: int = -1,
                    prefix_pattern='*[0-9]'):
        """PIV Plane initialized from a piv plane folder.

        Parameters
        ----------
        plane_directory: Union[str, pathlib.Path]
            Path to folder with snapshots
        time_info: Union[Tuple[datetime.datetime, float], List[datetime.datetime]]
            Time information based on which the relative and absolute time vectors are generated.
            The input can be two-fold:
            1. A Tuple of datetime object specifying the start of the PIV recording and a float specifying the
            measurement frequency
            2. A list of datetime objects corresponding to the recording datetime of each image
        pivfile: PIVFile
            The piv file object associated to a software.
        user_defined_hdf5_operations: Optional[Union[UserDefinedHDF5Operation, List[UserDefinedHDF5Operation]]], default=None
            User defined operations to be performed on the HDF5 file after the conversion.
        parameter: Union[PIVParameterInterface, str, pathlib.Path], optional=None
            If provided, this parameter interface will be used for all files.
            Default is None and searches for the parameter file associated to
            `pivfile`. However, in case the auto-detection finds multiple
            parameter files in the plane folder the process stops. This can
            be avoided by passing the parameter file explicitly here.
        n: int, default=-1
            Number of snapshot to take. Default value -1 means taking all found snapshots.
        prefix_pattern: str, default=*[0-9]?
            Pattern to match the snapshot files. Default is *[0-9]. The complete pattern is
            constructed as follows: prefix_pattern + pivfile.suffix

        Returns
        -------
        The initialized `PIVPlane` object
        """

        plane_directory = pathlib.Path(plane_directory)
        found_snapshot_files = scan_for_timeseries_files(plane_directory,
                                                         pivfile.suffix,
                                                         prefix_pattern=prefix_pattern)

        n_files = len(found_snapshot_files)
        logger.debug(f'Found {n_files} snapshot files here: {plane_directory.absolute()}')

        if n == -1:
            n = n_files
        found_snapshot_files = found_snapshot_files[0:n]
        logger.debug(f'Taking {len(found_snapshot_files)}/{n_files} snapshot')
        n_files = len(found_snapshot_files)
        if n_files == 0:
            raise ValueError(f'No snapshot files found in {plane_directory.absolute()}')

        pbar = tqdm(found_snapshot_files, desc=f'Init {n_files} {pivfile.__name__} object')
        if parameter:
            return PIVPlane([
                pivfile(
                    nc,
                    parameter=parameter,
                    user_defined_hdf5_operations=user_defined_hdf5_operations
                ) for nc in pbar],
                time_info)

        return PIVPlane([pivfile(nc, user_defined_hdf5_operations=user_defined_hdf5_operations) for nc in pbar],
                        time_info, )

    def __to_hdf__(self,
                   hdf_filename: Union[str, pathlib.Path] = None,
                   z: Union[float, str] = None,
                   **kwargs) -> pathlib.Path:
        """converts the snapshot into an HDF file

        Parameters
        ---------
        hdf_filename: Union[str, pathlib.Path], default=None
            The target hdf file. If None, a name is selected based on the plane folder name
        z: float, str, default=None
            the z-coordinate in [m] of the plane. Should be given if not present in the snapshot file or
            to be overwritten by this method. Instead of a float a string can be given, which is then
            interpreted as the z value with its unit, e.g. '1.2 mm'
        """

        iplane = kwargs.pop('iplane', None)
        nplanes = kwargs.pop('nplanes', None)
        title = kwargs.pop('title', DEFAULT_PLANE_TITLE)

        if hdf_filename is None:
            _parent = self.list_of_piv_files[0].filename.parent
            hdf_filename = _parent / f'{_parent.stem}.hdf'
        else:
            hdf_filename = pathlib.Path(hdf_filename)
            hdf_filename.parent.mkdir(parents=True, exist_ok=True)

        # get data from first snapshot to prepare the HDF5 file
        logger.debug('Reading first PIV file: '
                     f'{self.list_of_piv_files[0].__class__.__name__}({self.list_of_piv_files[0].filename}')
        pivdata = self.list_of_piv_files[0].read(self.rel_time_vector[0])

        if z is not None:
            pivdata.data['z'], z_units = parse_z(z)
        else:
            z_units = None
        mandatory_keys = ('x', 'y', 'z', 'ix', 'iy', 'reltime')
        for mkey in mandatory_keys:
            if mkey not in pivdata.data.keys():
                raise KeyError(f'Mandatory key {mkey} not provided.')
        nt, ny, nx = self.rel_time_vector.size, pivdata.data['y'].size, pivdata.data['x'].size
        _shape = dict(reltime=nt, y=ny, x=nx)
        dataset_shape = tuple([_shape[k] for k in self.plane_coord_order])
        _chunk = [_shape[n] for n in self.plane_coord_order]
        _chunk[self.plane_coord_order.index('reltime')] = 1
        dataset_chunk = tuple(_chunk)
        iy_idim = self.plane_coord_order.index('y')
        ix_idim = self.plane_coord_order.index('x')
        compression = get_config()['compression']
        compression_opts = get_config()['compression_opts']

        logger.debug(f'Writing plane data to (plane-) file {pathlib.Path(hdf_filename).absolute()}')

        if iplane is not None and nplanes is not None:
            pbar = tqdm(total=self.__len__(), desc=f'Writing plane HDF5 file {iplane}/{nplanes}')
        else:
            pbar = tqdm(total=self.__len__(), desc=f'Writing plane HDF5 file')

        with h5tbx.File(hdf_filename, 'w') as h5main:
            pbar.update()
            h5main.attrs['title'] = title
            for ak, av in pivdata.root_attrs.items():
                h5main.attrs[ak] = av
            # self.list_of_piv_files[0].write_parameters(h5main.create_group(PIV_PARAMETER_GRP_NAME))
            if PIV_PARAMETER_GRP_NAME not in h5main:
                h5main.create_group(PIV_PARAMETER_GRP_NAME)
            self.list_of_piv_files[0].write_parameters(h5main[PIV_PARAMETER_GRP_NAME])
            logger.debug('Creating datasets for coordinates (x, y, z, time, ix, iy)')

            if pivdata.data['ix'].min() < 0:
                raise ValueError('Data of ix must not be smaller than zero!')
            if pivdata.data['iy'].min() < 0:
                raise ValueError('Data of iy must not be smaller than zero!')

            for ds_name, maxshape in zip(('x', 'ix', 'y', 'iy', 'z'), (nx, nx, ny, ny, None)):
                h5main.create_dataset(name=ds_name,
                                      data=pivdata.data[ds_name],
                                      maxshape=maxshape,
                                      dtype=get_config()['dtypes'].get(ds_name, None))

            h5main.create_dataset('reltime', data=self.rel_time_vector,
                                  dtype=get_config()['dtypes'].get('reltime', None))

            for varkey in ('x', 'y', 'ix', 'iy', 'reltime'):
                h5main[varkey].make_scale()
                for ak, av in pivdata.data_attrs[varkey].items():
                    h5main[varkey].attrs[ak] = av

            if 'z' not in pivdata.data_attrs:
                h5main['z'].attrs['units'] = z_units
            else:
                for ak, av in pivdata.data_attrs['z'].items():
                    h5main['z'].attrs[ak] = av

            # write all other dataset to file:
            logger.debug('Writing abs time vector')
            ds_rec_dtime = create_recording_datetime_dataset(h5main, list(self.abs_time_vector), name='time')
            ds_rec_dtime.make_scale()
            ds_imgidx = h5main.create_dataset(IMAGE_INDEX, data=self.IMAGE_INDEX,
                                              attrs=dict(units='',
                                                         standard_name='piv_image_index'),
                                              dtype=get_config()['dtypes'].get(IMAGE_INDEX, None))
            ds_imgidx.make_scale()

            dataset_keys = []  # fill with dataset names that are not coordinates.

            for varkey, vardata in pivdata.data.items():
                if varkey not in ('x', 'y', 'ix', 'iy', 'z', 'reltime'):
                    dataset_keys.append(varkey)
                    ds = h5main.create_dataset(name=varkey,
                                               shape=dataset_shape,
                                               dtype=vardata.dtype,
                                               maxshape=dataset_shape,
                                               chunks=dataset_chunk,
                                               compression=compression,
                                               compression_opts=compression_opts)
                    for idim, coordname in enumerate(self.plane_coord_order):
                        ds.dims[idim].attach_scale(h5main[coordname])
                    ds.dims[iy_idim].attach_scale(h5main['iy'])
                    ds.dims[ix_idim].attach_scale(h5main['ix'])
                    ds.dims[0].attach_scale(ds_rec_dtime)
                    ds.dims[0].attach_scale(ds_imgidx)
                    ds.attrs['COORDINATES'] = ['z', ]
                    ds[0, ...] = vardata[...]

                    # write attributes to datasets
                    if varkey in pivdata.data_attrs:
                        for ak, av in pivdata.data_attrs[varkey].items():
                            ds.attrs[ak] = av

            logger.debug(f'Writing all datasets to file. Number of source files: {len(self.list_of_piv_files[1:])}')
            for (ifile, piv_file), t in zip(enumerate(self.list_of_piv_files[1:]), self.rel_time_vector[1:]):
                pbar.update(1)
                logger.debug(f'Processing {piv_file.filename}')
                _pivdata = piv_file.read(t)
                for varkey in dataset_keys:
                    h5main[varkey][ifile + 1, ...] = _pivdata.data[varkey][...]

                # pivview hack: flag meaning might be different for each plane because not all planes
                # have the same flag values and only those that appear are added to the dict
                if 'piv_flags' in _pivdata.data_attrs.keys():
                    try:
                        flag_meaning_plane = _pivdata.data_attrs['piv_flags']['flag_meaning']
                        if not isinstance(flag_meaning_plane, dict):
                            flag_meaning_plane_dict = json.loads(flag_meaning_plane)

                        try:
                            flag_meaning_main_dict = h5main['piv_flags'].attrs['flag_meaning']
                            if not isinstance(flag_meaning_main_dict, dict):
                                flag_meaning_main_dict = json.loads(flag_meaning_main_dict)
                            flag_meaning_main_dict.update(flag_meaning_plane_dict)
                            h5main['piv_flags'].attrs['flag_meaning'] = json.dumps(flag_meaning_main_dict)
                        except KeyError:
                            pass
                    except KeyError:
                        pass
        self.hdf_filename = hdf_filename
        return pathlib.Path(hdf_filename)


class PIVMultiPlane(PIVConverter):
    """Interface class"""
    __slots__ = 'list_of_piv_folder'
    plane_coord_order = ('z', 'reltime', 'y', 'x')

    def __getitem__(self, item):
        return self.list_of_piv_folder[item]

    def __init__(self, list_of_piv_planes: List[PIVPlane], ):
        super().__init__()
        if not isinstance(list_of_piv_planes, (tuple, list)):
            raise TypeError(
                f'Expecting a lists of list of {PIVPlane.__class__} objects but got {type(list_of_piv_planes)}')
        for i, piv_folder in enumerate(list_of_piv_planes):
            if len(piv_folder) == 0:
                raise ValueError(f'Plane number {i} is empty. Cannot proceed.')

        self.list_of_piv_planes = list_of_piv_planes
        pivfile_types = []
        ref_type = type(list_of_piv_planes[0].list_of_piv_files[0])
        for folder in list_of_piv_planes:
            pivfile_types.append(all(isinstance(f, ref_type) for f in folder.list_of_piv_files))
        if not all(pivfile_types):
            warnings.warn('The provided PIVFile objects are not all from the same class which may cause problems '
                          'in the process of converting them to HDF file(s).', UserWarning)

    def __len__(self):
        return len(self.list_of_piv_planes)

    def __repr__(self):
        out = f'<{self.__class__.__name__} of {self.__len__()} planes:\n\t'
        out += '\n\t'.join([p.__repr__() for p in self.list_of_piv_planes])
        out += '\n>'
        return out

    def __get_user_defined_hdf5_operations__(self) -> List[UserDefinedHDF5Operation]:
        return self.list_of_piv_planes[0].__get_user_defined_hdf5_operations__()

    @staticmethod
    def merge_planes(*,
                     hdf_filenames: List[pathlib.Path],
                     target_hdf_filename: Union[str, pathlib.Path],
                     title: str):
        """merges multiple piv plane hdf files together

        Parameters
        ----------
        hdf_filenames: List[pathlib.Path]
            List of hdf files to merge
        target_hdf_filename: Union[str, pathlib.Path]
            Name of the target hdf file
        title: str
            Title of the target hdf file
        """
        n_files = len(hdf_filenames)
        print(f'Merging {n_files} HDF files into {target_hdf_filename}.')
        software_list = []
        nt_list = []
        t_list = []
        x_data = []
        y_data = []
        z_data = []
        dim_names = []
        for hdf_file in hdf_filenames:
            with h5py.File(hdf_file) as h5plane:
                try:
                    software_list.append(h5plane.attrs['software'])
                except KeyError:
                    pass
                nt_list.append(h5plane['reltime'].size)
                t_list.append(h5plane['reltime'][()])
                x_data.append(h5plane['x'][:])
                y_data.append(h5plane['y'][:])
                z_data.append(h5plane['z'][()])
                dim_names.append([d[0].name for d in h5plane['u'].dims])
        if any(dim_names[0] != d for d in dim_names):
            raise RuntimeError('Inconsistent dimension names. All planes must have same dimension names'
                               'for all datasets.')
        # check compliance of planes:
        if len(software_list) > 1:
            if not np.all([software_list[0] == s for s in software_list[1:]]):
                raise ValueError('Plane have different software names registered')

        # they only can be merged if x and y data is equal but z is different:
        if not np.all([np.array_equal(x_data[0], x) for x in x_data[1:]]):
            raise ValueError('x coordinates of planes are different. Cannot merge.')
        if not np.all([np.array_equal(y_data[0], y) for y in y_data[1:]]):
            raise ValueError('y coordinates of planes are different. Cannot merge.')
        if np.any([z_data[0] == z for z in z_data[1:]]):
            raise ValueError(f'z coordinates must be different in order to merge the planes: {z_data}')

        with h5py.File(target_hdf_filename, 'w') as h5main:
            if software_list:
                h5main.attrs['software'] = software_list[0]
            # check if time vectors are close has been done by this time.
            nt_min = min(nt_list)
            dt_list = [np.diff(t[0:nt_min]) for t in t_list]
            abs_diff = np.abs(dt_list[0] - dt_list[1:])
            if np.all(abs_diff == 0.):
                print('Relative time vectors are identical. Merging into one group...')
                PIVMultiPlane._merge_planes_equal_time_vectors(h5main=h5main,
                                                               hdf_filenames=hdf_filenames,
                                                               title=title,
                                                               rel_time=np.asarray(t_list[0][0:nt_min]))
                return target_hdf_filename
            print('Time vectors are close enough but not identical. Merging into one group by using one '
                  'time vector. The time vectors of all planes are written in a separate group to keep them ...')
            av_time = np.asarray([t[:nt_min] for t in t_list]).mean(axis=0)
            PIVMultiPlane._merge_planes_equal_time_vectors(h5main=h5main,
                                                           hdf_filenames=hdf_filenames,
                                                           title=title,
                                                           rel_time=av_time)
            return target_hdf_filename

    @staticmethod
    def _merge_planes_equal_time_vectors(*,
                                         h5main: h5py.File,
                                         hdf_filenames: List[pathlib.Path],
                                         title: str,
                                         rel_time: [np.ndarray, List, None] = None) -> h5py.File:
        """

        If time is given, this time is used instead the one from the planes. Thus it is assumed
        that all plane times are close but not identical. Then time of each plane is
        put in a separate group.

        Parameters
        ----------
        h5main: h5py.File
            HDF file to write the merged data to
        hdf_filenames: List[pathlib.Path]
            List of hdf files to merge
        title: str
            Title of the target hdf file
        rel_time: [np.ndarray, List, None]=None
            Relative time vector to use.
        """
        config = get_config()
        nz = len(hdf_filenames)
        if rel_time is None:
            raise TypeError('rel_time must be given')

        if not isinstance(rel_time, np.ndarray):
            rel_time = np.asarray(rel_time)
        if not rel_time.ndim == 1:
            raise ValueError(f'rel_time must be 1D array but is {rel_time.ndim}D')

        nt = rel_time.size

        with h5py.File(hdf_filenames[0]) as h5plane:

            # write the absolute time of each plan into a dataset called "time", which will not be a dimension scale
            # for the PIV data!.
            ds_time = h5main.create_dataset('time', shape=(nz, nt), dtype=h5plane['time'].dtype)
            ds_time.attrs.update(h5plane['time'].attrs.items())
            ds_time[0, :] = h5plane['time'][:nt]
            for _iz, plane_hdf_filename in enumerate(hdf_filenames[1:]):
                with h5py.File(plane_hdf_filename) as _h5plane:
                    ds_time[_iz + 1, :] = _h5plane['time'][:nt]

            ny = h5plane['y'].size
            nx = h5plane['x'].size
            dim_names = [os.path.basename(d[0].name) for d in h5plane['u'].dims]
            plane_coord_order = PIVMultiPlane.plane_coord_order
            shape_dict = {'reltime': nt, 'x': nx, 'y': ny}
            _ds_shape_list = [shape_dict[n] for n in dim_names]

            ix = plane_coord_order.index('x')
            iy = plane_coord_order.index('y')
            if abs(ix - iy) > 1:
                raise ValueError('Invalid plane coord position. x and y must be next to each other.')
            zidx = plane_coord_order.index('z')
            it = plane_coord_order.index('reltime')
            if zidx == 0 and it == 1:
                ds_shape = (nz, nt, ny, nx)
                ds_chunk = (1, 1, ny, nx)
            elif zidx == 1 and it == 0:
                ds_shape = (nt, nz, ny, nx)
                ds_chunk = (1, 1, ny, nx)
            else:
                raise NotImplementedError('Cannot work with that shape...')

            coord_names = ('x', 'y', 'reltime', 'z', 'ix', 'iy', 'iz', 'time')
            dataset_properties = {k: {'dtype': v.dtype} for k, v in h5plane.items() if
                                  isinstance(v, h5py.Dataset) and k not in coord_names and v.ndim > 1}

            compression = h5plane['u'].compression
            compression_opts = h5plane['u'].compression_opts

            x = h5plane['x'][:]
            y = h5plane['y'][:]
            ix = h5plane['ix'][:]
            iy = h5plane['iy'][:]

        iz = np.arange(1, nz + 1, 1, dtype=int)

        h5main.attrs['title'] = title
        ds_x = h5main.create_dataset('x', data=x, dtype=config['dtypes'].get('x', None))
        ds_x.make_scale()
        ds_y = h5main.create_dataset('y', data=y, dtype=config['dtypes'].get('y', None))
        ds_y.make_scale()

        ds_ix = h5main.create_dataset('ix', data=ix, dtype=config['dtypes'].get('ix', get_uint_type(ix)))
        ds_ix.make_scale()
        ds_iy = h5main.create_dataset('iy', data=iy, dtype=config['dtypes'].get('iy', get_uint_type(iy)))
        ds_iy.make_scale()
        ds_iz = h5main.create_dataset('iz', data=iz, dtype=config['dtypes'].get('iz', get_uint_type(iz)))
        ds_iz.make_scale()

        ds_z = h5main.create_dataset('z', shape=(nz,), dtype=config['dtypes'].get('z', None))
        ds_z.make_scale()
        ds_relt = h5main.create_dataset('reltime', shape=(nt,), dtype=config['dtypes'].get('reltime', None))
        ds_relt[:] = rel_time
        ds_relt.make_scale()

        ds_imgidx = h5main.create_dataset(IMAGE_INDEX, shape=(nt,), dtype=config['dtypes'].get(IMAGE_INDEX, None))
        ds_imgidx[:] = np.arange(1, nt + 1, 1, dtype=int)
        ds_imgidx.make_scale()

        # ds_time.dims[0].attach_scale(h5main['z'])
        # ds_time.dims[1].attach_scale('reltime')
        for ds_name, ds_props in dataset_properties.items():
            ds = h5main.create_dataset(ds_name,
                                       shape=ds_shape,
                                       chunks=ds_chunk,
                                       dtype=ds_props['dtype'],
                                       compression=compression,
                                       compression_opts=compression_opts)
            for i, n in enumerate(plane_coord_order):
                ds.dims[i].attach_scale(h5main[n])
                if n == 'z':
                    ds.dims[i].attach_scale(h5main['iz'])
                if n == 'reltime':
                    # ds.dims[i].attach_scale(h5main['time'])  # don't attach, because it is 2D! and anyhow a common dimension cannot be found
                    ds.dims[i].attach_scale(h5main[IMAGE_INDEX])
                if n == 'x':
                    ds.dims[i].attach_scale(h5main['ix'])
                if n == 'y':
                    ds.dims[i].attach_scale(h5main['iy'])

        with h5py.File(hdf_filenames[0]) as h5plane:

            for ds_name in dataset_properties.keys():
                for ak, av in h5plane[ds_name].attrs.items():
                    if not ak.isupper():
                        h5main[ds_name].attrs[ak] = av

            # if time is not None:
            #     # time has been set already. now put all plane times in a separate group
            #     plane_times_group = h5main.create_group(PLANE_TIME_GROUP_NAME)

            for n in ('x', 'y', 'ix', 'iy', 'z', 'reltime'):
                for ak, av in h5plane[n].attrs.items():
                    if not ak.isupper():
                        h5main[n].attrs[ak] = av

        z_before_t = zidx < it

        for iplane, plane_hdf_filename in enumerate(hdf_filenames):
            with h5py.File(plane_hdf_filename) as h5plane:
                # if time is not None:
                # ds_plane_time = plane_times_group.create_dataset(f'planetime_{iplane:0{len(str(nt))}d}',
                #                                                  data=h5plane['reltime'][:],
                #                                                  dtype=config['dtypes'].get('reltime', None))
                # for ak, av in h5plane['reltime'].attrs.items():
                #     if not ak.isupper():
                #         ds_plane_time.attrs[ak] = av

                h5main['z'][iplane] = h5plane['z'][()]
                for k in dataset_properties.keys():
                    ds = h5main[k]
                    if z_before_t:
                        ds[iplane, :] = h5plane[k][0:nt, ...]
                    else:
                        for _it in range(nt):
                            ds[_it, iplane, ...] = h5plane[k][_it, :, :]
                # write piv parameter group
                plane_grp = h5main.create_group(f'plane{iplane:0{len(str(nz))}}')
                if PIV_PARAMETER_GRP_NAME not in plane_grp:
                    trg_grp = plane_grp.create_group(PIV_PARAMETER_GRP_NAME)
                else:
                    trg_grp = plane_grp[PIV_PARAMETER_GRP_NAME]
                copy_piv_parameter_group(h5plane[PIV_PARAMETER_GRP_NAME], trg_grp)

        return h5main

    @staticmethod
    def _merge_planes_unequal_time_vectors(h5main: h5py.File, hdf_filenames: List[pathlib.Path]) -> None:
        """merges multiple hdf files that have different time vectors lengths or entries"""
        nt_list = []
        x_data = []
        y_data = []
        z_data = []
        nz = len(hdf_filenames)
        dim_names = []
        for hdf_file in hdf_filenames:
            with h5py.File(hdf_file) as h5plane:
                nt_list.append(h5plane['reltime'].size)
                x_data.append(h5plane['x'][:])
                y_data.append(h5plane['y'][:])
                z_data.append(h5plane['z'][()])
                dim_names.append([d[0].name for d in h5plane['u'].dims])
        # different time steps --> put data in individual groups
        plane_grps = []

        h5main.attrs['title'] = DEFAULT_MPLANE_TITLE
        for iz, hdf_file in tqdm(enumerate(hdf_filenames), total=nz, desc='Copying groups to target file'):
            plane_grp = h5main.create_group(f'plane{iz:0{len(str(nz))}}')
            plane_grps.append(plane_grp)
            with h5py.File(hdf_file) as h5plane:
                for objname in h5plane:
                    if objname != PIV_PARAMETER_GRP_NAME:  # treat separately
                        h5main.copy(h5plane[objname], plane_grp)
                        # delete dimension scale attrs:
                        for ak in ('DIMENSION_LIST', 'REFERENCE_LIST', 'CLASS', 'NAME'):
                            try:
                                del plane_grp[objname].attrs[ak]
                            except KeyError:
                                pass
                    else:
                        copy_piv_parameter_group(h5plane[PIV_PARAMETER_GRP_NAME],
                                                 plane_grp.create_group(PIV_PARAMETER_GRP_NAME))

        for varkey in ('x', 'y', 'ix', 'iy'):
            h5main.move(plane_grps[0][varkey].name, varkey)
            h5main[varkey].make_scale()
            for plane_grp in plane_grps[1:]:
                del plane_grp[varkey]

        for plane_grp in plane_grps[1:]:
            _time_ds = plane_grp['reltime']
            for ak in ('DIMENSION_LIST', 'REFERENCE_LIST', 'CLASS', 'NAME'):
                try:
                    del _time_ds.attrs[ak]
                except KeyError:
                    pass
            _time_ds.make_scale()

        piv_variables = []
        for plane_grp in plane_grps[:]:
            for k, v in plane_grp.items():
                if isinstance(v, h5py.Dataset):
                    if v.ndim > 1:
                        piv_variables.append(v)
                        for i, d in enumerate(dim_names[0]):
                            if 'reltime' in d and 'reltime' in plane_grp:
                                # time is stored in each group
                                v.dims[i].attach_scale(plane_grp['reltime'])
                            else:
                                # coordinates one group above (in root)
                                v.dims[i].attach_scale(h5main[d])

        # sanity check:
        for piv_variable in piv_variables:
            if piv_variable.dims[0][0] != piv_variable.parent['reltime']:
                raise ValueError('Time dimension is not the first dimension of the dataset.')
            if piv_variable.dims[1][0] != h5main['y']:
                raise ValueError('y dimension is not the second dimension of the dataset.')
            if piv_variable.dims[2][0] != h5main['x']:
                raise ValueError('x dimension is not the third dimension of the dataset.')

    @staticmethod
    def from_folders(plane_directories: Union[List[str], List[pathlib.Path]],
                     *,
                     time_infos: List[Union[Tuple[datetime.datetime, float], List[datetime.datetime]]],
                     pivfile: Type[PIVFile],
                     user_defined_hdf5_operations: Optional[
                         Union[UserDefinedHDF5Operation, List[UserDefinedHDF5Operation]]
                     ] = None,
                     n: int = -1) -> "PIVMultiPlane":
        """init PIVPlane from multiple folders

        Parameter
        ----------
        plane_directories : list of str or pathlib.Path
            list of directories containing the plane data
        time_infos : List[Union[Tuple[datetime.datetime, float], List[datetime.datetime]]]+
            List of time_info. See PIVPlane docstring for more information about time_info
        pivfile : PIVFile
            PIVFile object
        n : int, default=-1
            number of snapshots to load. If n=-1, all snapshots are loaded
        """
        if len(time_infos) != len(plane_directories):
            raise ValueError("Number of planes don't match the number of recording (time) information")
        plane_objs = [PIVPlane.from_folder(pathlib.Path(d),
                                           time_info=time_info,
                                           pivfile=pivfile,
                                           user_defined_hdf5_operations=user_defined_hdf5_operations,
                                           n=n) for d, time_info
                      in
                      zip(plane_directories, time_infos)]
        return PIVMultiPlane(plane_objs)

    def __to_hdf__(self,
                   hdf_filename: Union[str, pathlib.Path] = None,
                   rtol: float = 1.e-5,
                   atol: float = 1.e-8,
                   trim_time_vectors: bool = True,
                   z: Union[List[float], np.ndarray] = None,
                   **kwargs) -> pathlib.Path:
        """converts the snapshot into an HDF file"""
        title = kwargs.pop('title', DEFAULT_MPLANE_TITLE)

        # if the time vectors are known, check if planes can be merged!
        t_list = [plane.rel_time_vector for plane in self.list_of_piv_planes]
        nt_list = [len(t) for t in t_list]
        equal_length = np.all([nt_list[0] == nt for nt in nt_list[1:]])
        # if not equal_length:
        #     # cannot merge into one dataset. put each plane into a separate group or trim the vectors to the common minimal length
        if not equal_length and not trim_time_vectors:
            raise ValueError(
                'Time vectors are not equal in length! Stopping here because "trim_time_vectors" is False. '
                'Set "trim_time_vectors" to True to trim the time vectors to the shortest one.'
            )

        # check if time vectors are close:
        nt_min = min(nt_list)

        # compute time deltas:
        dt_list = [np.diff(t[:nt_min]) for t in t_list]

        if atol is not None:
            abs_diff = [np.abs(dt_list[0] - dt) for dt in dt_list[1:]]
            atol_result = np.all([a < atol for a in abs_diff])
        else:
            atol_result = True
        if rtol is not None:
            rel_diff = [np.abs((dt_list[0] - dt) / dt_list[0]) for dt in dt_list[1:]]
            rtol_result = np.all([r < rtol for r in rel_diff])
        else:
            rtol_result = True

        if not rtol_result or not atol_result:
            raise ValueError('Time vectors are not close enough to be merged! Process each plane separately.')

        if hdf_filename is None:
            snapshot0_filename = self.list_of_piv_planes[0].list_of_piv_files[0].filename
            name = f'{snapshot0_filename.parent.parent}.hdf'
            hdf_filename = snapshot0_filename.parent.parent / name

        # get data from first snapshot to prepare the HDF5 file
        nplanes = len(self.list_of_piv_planes)
        if z is None:
            z = [None, ] * nplanes
        plane_hdf_files = [plane.__to_hdf__(generate_temporary_filename(suffix='_plane.hdf'),
                                            z=_z,
                                            iplane=iplane + 1,
                                            nplanes=nplanes) for (iplane, plane), _z
                           in zip(enumerate(self.list_of_piv_planes), z)]
        hdf_filename = self.merge_planes(hdf_filenames=plane_hdf_files,
                                         target_hdf_filename=hdf_filename,
                                         title=title)
        print('Done.')
        return hdf_filename

    def to_hdf(self,
               piv_attributes: Dict,  # e.g. piv_medium, contact
               hdf_filename: pathlib.Path = None,
               z: float = None,
               atol: float = 1e-6,
               rtol: float = 1e-3) -> pathlib.Path:
        return super().to_hdf(piv_attributes=piv_attributes,
                              hdf_filename=hdf_filename,
                              z=z,
                              atol=atol,
                              rtol=rtol)



import ast
import configparser
import pathlib
from ast import literal_eval

import h5py
import pint
from netCDF4 import Dataset as ncDataset

from ..interface import PIVParameterInterface


def is_nc_file(nc_filename: pathlib.Path):
    """try to open a nc file and returns success"""
    try:
        with ncDataset(nc_filename, 'r') as nc_rootgrp:
            _ = nc_rootgrp.dimensions
            return True
    except OSError:
        return False


def eval_piv_param(value):
    try:
        return literal_eval(value)
    except Exception:
        # print(f'Could not evaluate value: {value}. Returning as string.')
        return value


class PIVviewParamFile(PIVParameterInterface):
    """Parameter file interface for PIVview"""

    __suffix__ = '.par'

    def __init__(self, filename):
        super().__init__(filename)
        self._read_file()

    def _read_file(self):
        _cfg = configparser.ConfigParser()
        _cfg.read(self.filename)
        self.param_dict = {}
        for s in _cfg.sections():
            self.param_dict[s.strip('-').strip(' ')] = {k: eval_piv_param(v) for k, v in dict(_cfg[s]).items()}
        # fill the rest of the parameters
        piv_proc_param = self.param_dict['PIV processing parameters']
        x_final_iw_size, y_final_iw_size, _ = piv_proc_param['view0_piv_eval_samplesize']
        self.piv_param_datasets['x_final_iw_size']['data'] = x_final_iw_size
        self.piv_param_datasets['y_final_iw_size']['data'] = y_final_iw_size
        x_final_iw_overlap_size, y_final_iw_overlap_size, _ = piv_proc_param['view0_piv_eval_samplestep']
        self.piv_param_datasets['x_final_iw_overlap_size']['data'] = x_final_iw_overlap_size
        self.piv_param_datasets['y_final_iw_overlap_size']['data'] = y_final_iw_overlap_size

        dt = self.param_dict['PIV conversion parameters']['view0_piv_conv_pulsedelay']
        dt_unit = self.param_dict['PIV conversion parameters']['view0_piv_conv_pulsedelayunits']
        self.piv_param_datasets['laser_pulse_delay']['data'] = dt
        self.piv_param_datasets['laser_pulse_delay']['units'] = dt_unit

        cf = self.param_dict['PIV conversion parameters']['view0_piv_conv_lengthconversion']
        cf_units = f"pixel/{self.param_dict['PIV conversion parameters']['view0_piv_conv_lengthconversionunits']}"
        cf_q = pint.Quantity(f'{cf} {cf_units}').to('pixel/m')
        self.piv_param_datasets['piv_scaling_factor']['data'] = cf_q.magnitude

    def save(self, filename: pathlib.Path):
        """Save to original file format"""
        _cfg = configparser.ConfigParser()
        _cfg.read_dict(self.param_dict)
        with open(filename, 'w') as f:
            self._cfg.write(f)

    def from_hdf(self, grp: h5py.Group) -> None:
        """Read from HDF group"""
        pass


class PivViewConfigFile(PIVviewParamFile):
    """Parameter/Config file for stereo PIVview files"""
    __suffix__ = '.cfg'

    def _read_file(self):
        _cfg = configparser.ConfigParser()
        _cfg.read(self.filename)
        self.param_dict = {}
        for s in _cfg.sections():
            dict_holder = {}
            for k, v in dict(_cfg[s]).items():
                try:
                    dict_holder[k] = ast.literal_eval(v)
                except SyntaxError:
                    dict_holder[k] = v
            self.param_dict[s.strip('-').strip(' ')] = dict_holder

        piv_proc_param = self.param_dict['PIV processing parameters']
        x_final_iw_size, y_final_iw_size, _ = piv_proc_param['view0_piv_eval_samplesize']
        self.piv_param_datasets['x_final_iw_size']['data'] = x_final_iw_size
        self.piv_param_datasets['y_final_iw_size']['data'] = y_final_iw_size
        x_final_iw_overlap_size, y_final_iw_overlap_size, _ = piv_proc_param['view0_piv_eval_samplestep']
        self.piv_param_datasets['x_final_iw_overlap_size']['data'] = x_final_iw_overlap_size
        self.piv_param_datasets['y_final_iw_overlap_size']['data'] = y_final_iw_overlap_size

        dt = self.param_dict['PIV conversion parameters']['view0_piv_conv_pulsedelay']
        dt_unit = self.param_dict['PIV conversion parameters']['view0_piv_conv_pulsedelayunits']
        self.piv_param_datasets['laser_pulse_delay']['data'] = dt
        self.piv_param_datasets['laser_pulse_delay']['units'] = dt_unit

        cf = self.param_dict['PIV conversion parameters']['view0_piv_conv_lengthconversion']
        cf_units = f"pixel/{self.param_dict['PIV conversion parameters']['view0_piv_conv_lengthconversionunits']}"
        cf_q = pint.Quantity(f'{cf} {cf_units}').to('pixel/m')
        self.piv_param_datasets['piv_scaling_factor']['data'] = cf_q.magnitude

        assert len(self.param_dict) != 0, f'Could not read config file: {self.filename}'



import json
import pathlib
import unittest

import h5rdmtoolbox as h5tbx
from h5rdmtoolbox import convention

import piv2hdf

__this_dir__ = pathlib.Path(__file__).parent


def read_meta():
    with open(__this_dir__ / "resources/meta.json") as f:
        meta = json.load(f)
    return meta


class TestConfig(unittest.TestCase):

    def test_contacts(self):
        contacts_bak = piv2hdf.contacts
        contacts = piv2hdf.contacts
        contacts.clear()
        self.assertEqual(len(contacts), 0)
        self.assertTrue(contacts.__repr__().startswith('Contacts('))
        self.assertEqual(len(contacts.registered_contacts), len(contacts))
        contacts.register(mp='https://orcid.org/0000-0001-8729-0482')
        self.assertDictEqual({'mp': 'https://orcid.org/0000-0001-8729-0482', },
                             contacts.registered_contacts)
        contacts.register(mp2='https://orcid.org/0000-0001-8729-0482')
        self.assertDictEqual({'mp': 'https://orcid.org/0000-0001-8729-0482',
                              'mp2': 'https://orcid.org/0000-0001-8729-0482'}, contacts.registered_contacts)

        # print(contacts.contacts_filename)
        self.assertEqual(len(contacts.registered_contacts), len(contacts))
        contacts.contacts_filename.unlink()

        _contacts = piv2hdf.Contacts()
        self.assertDictEqual({}, _contacts.registered_contacts)

        contacts.registered_contacts = {}
        contacts.register(mp='https://orcid.org/0000-0001-8729-0482')
        self.assertDictEqual({'mp': 'https://orcid.org/0000-0001-8729-0482'}, contacts.registered_contacts)

        contacts.clear()
        contacts.register(mp='https://orcid.org/0000-0001-8729-0482')
        self.assertDictEqual({'mp': 'https://orcid.org/0000-0001-8729-0482'}, contacts.registered_contacts)

        contacts_bak.save()

    def test_set_config(self):
        package_convention = piv2hdf.CONVENTION_FILENAME
        cv = h5tbx.Convention.from_yaml(package_convention)
        cv.register()

        current_cv_name = piv2hdf.get_config('convention')
        self.assertEqual(current_cv_name, current_cv_name)

        piv2hdf.set_config(convention=None)
        current_cv_name = convention.get_current_convention().name
        self.assertEqual('h5py', current_cv_name)

        piv2hdf.set_config(convention='h5tbx')
        current_cv_name = convention.get_current_convention().name
        self.assertEqual('h5py', current_cv_name)

        piv2hdf.set_config(convention=cv)
        current_cv_name = convention.get_current_convention().name
        self.assertEqual('h5py', current_cv_name)

        with piv2hdf.set_config(convention='h5py'):
            current_cv_name = convention.get_current_convention().name
            self.assertEqual('h5py', current_cv_name)
            self.assertEqual('h5py', piv2hdf.get_config('convention'))

        current_cv_name = convention.get_current_convention().name
        self.assertEqual('h5py', current_cv_name)
        self.assertEqual(cv, piv2hdf.get_config('convention'))

    def test_set_pivattrs(self):
        meta = read_meta()
        _ = piv2hdf.set_pivattrs(creator=meta["CREATOR"],
                                 piv_medium='air',
                                 camera_type='My Camera',
                                 fstop=1.0,
                                 laser={'name': 'Big Sky laser', 'energy per pulse:': '130 mJ'})



import unittest

from piv2hdf import tutorial
from piv2hdf.openpiv import OpenPIVParameterFile
from piv2hdf.pivview import PIVviewParamFile


class TestParameter(unittest.TestCase):

    def test_pivview_parameter(self):
        with self.assertRaises(FileNotFoundError):
            PIVviewParamFile('nonexistent.par')
        pivview_parameter_file = tutorial.PIVview.get_parameter_file()
        par = PIVviewParamFile(pivview_parameter_file)
        self.assertEqual(par.suffix, '.par')
        pivview_parameter_dict = par.to_dict()
        self.assertIsInstance(pivview_parameter_dict, dict)

    def test_openpiv_parameter(self):
        openpiv_parameter_file = tutorial.OpenPIV.get_parameter_file()
        par = OpenPIVParameterFile(openpiv_parameter_file)
        self.assertEqual(par.suffix, '.par')
        openpiv_parameter_dict = par.to_dict()
        self.assertIsInstance(openpiv_parameter_dict, dict)



import pathlib
import unittest

import h5rdmtoolbox as h5tbx

from piv2hdf import cv


class TestConvention(unittest.TestCase):

    def test_snt_filename_exists(self):
        self.assertTrue(
            pathlib.Path(pathlib.Path(cv.properties[h5tbx.File]['standard_name_table'].default_value.value)).exists())



"""consts for PIVview module"""

from os import environ

from pathlib import Path

PIVVIEW_SOFTWARE = dict(name='pivview', version=999,
                        url='https://www.pivtec.com',
                        description='PIVTEC PIVview')

DIM_NAMES = ('z', 'reltime', 'y', 'x', 'ix', 'iy')
DIM_NAMES_TIMEAVERAGED = ('z', 'y', 'x', 'ix', 'iy')
DEFAULT_DATASET_LONG_NAMES = {'x': 'x-coordinate',
                              'y': 'y-coordinate',
                              'z': 'z-coordinate',
                              'ix': 'x-pixel-coordinate',
                              'iy': 'y-pixel-coordinate',
                              'reltime': 'reltime'}
IGNORE_ATTRS = ('mean_dx', 'mean_dy', 'mean_dz',
                'rms_dx', 'rms_dy', 'rms_dz',
                'coord_min', 'coord_max', 'coord_units', 'time_step', 'time_step_idx',
                'CLASS', 'NAME', 'DIMENSION_LIST', 'REFERENCE_LIST', 'COORDINATES')

MULTIPLANE_TITLE = 'PIV multi-plane file generated from PIVview netCDF4 files.'
PIVPLANE_TITLE = 'PIV plane file generated from PIVview netCDF4 files.'
PIVSNAPSHOT_TITLE = 'PIV snapshot file generated from a single PIVview netCDF4 file.'

PIV_FILE_TYPE_NAME = {'PIVSnapshot': 'snapshot',
                      'PIVPlane': 'single_plane',
                      'PIVMultiPlane': 'multi_plane'}

try:
    NCDF2PAR = Path(environ.get('ncdf2par'))
except TypeError:
    NCDF2PAR = None



import logging
import pathlib
from logging.handlers import RotatingFileHandler
from typing import Union

import appdirs

name = __package__

_logdir = appdirs.user_log_dir(name)
_log = pathlib.Path(_logdir)  # Currently, this is unversioned

try:
    pathlib.Path.mkdir(_log, parents=True)
    _logFolderMsg = f'{name} log folder created at {_logdir}'
except FileExistsError:
    _logFolderMsg = f'{name} log folder available: {_logdir}'

# Initialize logger, set high level to prevent ipython debugs. File level is
# set below
DEFAULT_LOGGING_LEVEL = logging.INFO
logger = logging.getLogger(__package__)
logger.setLevel(DEFAULT_LOGGING_LEVEL)

_formatter = logging.Formatter(
    '%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
    datefmt='%Y-%m-%d_%H:%M:%S')

_file_handler = RotatingFileHandler(_log / f'{name}.log', maxBytes=int(5e6), backupCount=2)
_file_handler.setLevel(DEFAULT_LOGGING_LEVEL)
_file_handler.setFormatter(_formatter)

_stream_handler = logging.StreamHandler()
_stream_handler.setLevel(DEFAULT_LOGGING_LEVEL)
_stream_handler.setFormatter(_formatter)

logger.addHandler(_file_handler)
logger.addHandler(_stream_handler)

# Log messages collected above
logger.debug(_logFolderMsg)


def set_loglevel(level: Union[str, int]):
    """Set the log level"""
    logger.setLevel(level)
    for handler in logger.handlers:
        handler.setLevel(level)



"""tests for snapshot conversion"""
import datetime
import json
import logging
import pathlib
import unittest

import h5rdmtoolbox as h5tbx
import numpy as np

import piv2hdf
from piv2hdf import openpiv, pivview, tutorial, PIVSnapshot
from piv2hdf import set_loglevel, logger
from piv2hdf.interface import PIVFile
from piv2hdf.openpiv.user_operations import add_standard_name_operation as openpiv_add_standard_name_operation
from piv2hdf.pivview.user_operations import add_standard_name_operation as pivview_add_standard_name_operation

loglevel = logging.ERROR

set_loglevel(loglevel)
assert logger.level == loglevel
for handler in logger.handlers:
    assert handler.level == loglevel

__this_dir__ = pathlib.Path(__file__).parent


def read_meta():
    with open(__this_dir__ / "resources/meta.json") as f:
        meta = json.load(f)
    return meta


try:
    import lvpyio as lv

    lvpyio_installed = True
except ImportError:
    lvpyio_installed = False


class TestSnapshot(unittest.TestCase):
    if lvpyio_installed:
        def test_davis_snapshot(self):
            from piv2hdf.davis import VC7File
            from piv2hdf.davis.parameter import DavisParameterFile
            vc7_filename = tutorial.Davis.get_vc7_files()[1]
            self.assertTrue(vc7_filename.exists())
            param = DavisParameterFile(vc7_filename)
            vc7file = VC7File(vc7_filename, parameter=param)
            self.assertIsInstance(vc7file, PIVFile)
            self.assertEqual(vc7file.filename, vc7_filename)
            snapshot = PIVSnapshot(
                piv_file=vc7file,
                recording_dtime=datetime.datetime(2023, 1, 15, 13, 42, 2, 3)
            )
            snapshot.to_hdf()

    def test_pivview_snapshot(self):
        meta = read_meta()

        pivview_nc_file = tutorial.PIVview.get_snapshot_nc_files()[0]
        pivfile = pivview.PIVViewNcFile(
            filename=pivview_nc_file,
            parameter_filename=None,
            user_defined_hdf5_operations=pivview_add_standard_name_operation
        )
        snapshot = PIVSnapshot(piv_file=pivfile,
                               recording_dtime=datetime.datetime(2023, 1, 15, 13, 42, 2, 3))
        self.assertIsInstance(pivfile, PIVFile)

        piv2hdf.reset_pivattrs()

        # with self.assertRaises(h5tbx.errors.StandardAttributeError):
        #     snapshot.to_hdf()

        with self.assertRaises(TypeError):
            PIVSnapshot(piv_file=pivfile, recording_dtime=[0., 1.])

        snapshot = PIVSnapshot(piv_file=pivfile,
                               recording_dtime=datetime.datetime(2023, 1, 15, 13, 42, 2, 3))
        # with self.assertRaises(h5tbx.errors.StandardAttributeError):
        #     _ = snapshot.to_hdf(piv_attributes={'piv_medium': meta["PIV_MEDIUM"]})

        hdf_filename = snapshot.to_hdf(piv_attributes={'creator': meta["CREATOR"],
                                                       'piv_medium': meta["PIV_MEDIUM"],
                                                       'camera': meta["CAMERA"]})
        self.assertEqual(hdf_filename, snapshot.hdf_filename)
        nc_data = snapshot.piv_file.read(0).data

        with h5tbx.File(hdf_filename) as h5:
            self.assertEqual(h5.attrs['creator'], meta["CREATOR"])
            self.assertEqual(h5.attrs['piv_medium'], meta["PIV_MEDIUM"])
            self.assertEqual(h5.attrs['camera'], meta["CAMERA"])
            for k in nc_data.keys():
                if k not in ('z', 'reltime'):
                    self.assertIn(k, h5)
                    self.assertEqual(h5[k].shape, nc_data[k].shape)
                    self.assertEqual(h5[k].dtype, nc_data[k].dtype)
                    np.testing.assert_array_equal(h5[k][()], nc_data[k][()])
            self.assertIn('time', h5)
            self.assertIn('piv_medium', h5.attrs.raw)
            self.assertIn('piv_method', h5.piv_parameters.attrs.raw)
            self.assertIn('piv_peak_method', h5.piv_parameters.attrs.raw)
            self.assertIn('creator', h5.attrs.raw)
            self.assertIn('software', h5.attrs)
            # self.assertIn('recording_datetime', h5)  # datetime of plane (or in this case snapshot)
            self.assertEqual(h5['x'].size, 31)
            self.assertEqual(h5['y'].size, 15)
            self.assertEqual(h5['z'].size, 1)
            self.assertEqual(str(h5['u'][()].time.values, encoding='utf-8'),
                             datetime.datetime(2023, 1, 15, 13, 42, 2, 3).isoformat())
            self.assertIn('time', h5['u'][()].coords)
            self.assertEqual(h5['u'].shape, (15, 31))
            self.assertEqual(h5['v'].shape, (15, 31))
            self.assertIsNotNone(h5.find_one({'standard_name': {'$regex': 'piv_flags?'}}))
            self.assertEqual('uint8',
                             h5.find_one({'standard_name': {'$regex': 'piv_flags?'}}).dtype)

    def test_openpiv_snapshot(self):
        # get the test file:
        meta = read_meta()
        openpiv_txt_file = tutorial.OpenPIV.get_snapshot_txt_file()

        # init a openpiv-file instance:
        # None -> auto search
        openpiv_file = openpiv.OpenPIVFile(
            openpiv_txt_file,
            user_defined_hdf5_operations=openpiv_add_standard_name_operation,
            parameter_filename=None)

        openpiv_snapshot = PIVSnapshot(openpiv_file, recording_dtime=None)

        hdf_filename = openpiv_snapshot.to_hdf(piv_attributes={'creator': meta["CREATOR"],
                                                               'piv_medium': meta["PIV_MEDIUM"]})

        with h5tbx.File(hdf_filename) as h5:
            self.assertIn('piv_medium', h5.attrs.raw)
            # self.assertIn('piv_method', h5.piv_parameters.attrs.raw)
            self.assertIn('piv_peak_method', h5.piv_parameters.attrs.raw)
            self.assertIn('software', h5.attrs)
            self.assertIn('reltime', h5)
            self.assertEqual(h5['x'].size, 63)
            self.assertEqual(h5['y'].size, 59)
            self.assertEqual(h5['u'].shape, (59, 63))
            self.assertEqual(h5['v'].shape, (59, 63))
            self.assertIsNotNone(h5.find_one({'standard_name': {'$regex': 'piv_flag'}}))
            self.assertEqual(h5.find_one({'standard_name': {'$regex': 'piv_flag'}}).dtype, 'uint8')



import pathlib

import h5rdmtoolbox as h5tbx

from piv2hdf.interface import PIV_PARAMETER_GRP_NAME
from piv2hdf.interface import UserDefinedHDF5Operation
from piv2hdf.piv_params import PIV_PeakFitMethod
from piv2hdf.utils import read_translation_yaml_file

__this_dir__ = pathlib.Path(__file__).parent
RESOURCES_DIR = (__this_dir__.parent / '../resources').resolve()
TRANSLATION_EXT_DICT = read_translation_yaml_file(RESOURCES_DIR / 'openpiv/openpiv_ext_translation.yaml')


class AddStandardNameOperation(UserDefinedHDF5Operation):

    def __call__(self, h5: h5tbx.File) -> None:
        """openpiv post function"""
        for name, ds in h5.items():
            if name in TRANSLATION_EXT_DICT:
                ds.attrs['standard_name'] = TRANSLATION_EXT_DICT[name]

        def _update_fields(grp):
            peak_method = grp.attrs['subpixel_method']
            if peak_method == 'gaussian':
                grp.attrs['piv_peak_method'] = PIV_PeakFitMethod(0).name

            # # piv_method
            # warnings.warn('piv_method is assumed to be multi grid but not determined!')
            # grp.attrs['piv_method'] = PIV_METHOD(2).name

        if PIV_PARAMETER_GRP_NAME not in h5:
            for param_grp in h5.find({'$basename': PIV_PARAMETER_GRP_NAME}, recursive=True):
                _update_fields(param_grp)
            return

        return _update_fields(h5[PIV_PARAMETER_GRP_NAME])


add_standard_name_operation = AddStandardNameOperation()



"""User specific constants such as paths"""

import pathlib

import appdirs

from ._version import __version__

USER_DIRECTORY = pathlib.Path(appdirs.user_data_dir('piv2hdf', version=__version__))

TEST_DATA_DIRECTORY = pathlib.Path(__file__).parent.parent / 'tests/resources'
TEMPORARY_USER_DIRECTORY = USER_DIRECTORY / 'tmp'
TEMPORARY_USER_DIRECTORY.mkdir(parents=True, exist_ok=True)
USER_DIRECTORY.mkdir(parents=True, exist_ok=True)



import pathlib

import h5rdmtoolbox as h5tbx
from ssnolib.namespace import SSNO

from piv2hdf.interface import PIV_PARAMETER_GRP_NAME, PIV_PARAMETER_ATTRS_NAME
from piv2hdf.piv_params import PIV_PeakFitMethod, PIV_METHOD
from piv2hdf.utils import read_translation_yaml_file

__this_file__ = pathlib.Path(__file__).resolve()

_RESOURCES = __this_file__.parent / '../../resources'
PIVVIEW_TRANSLATION = read_translation_yaml_file(_RESOURCES / 'pivview/pivview_translation.yaml')


class AddStandardNameOperation:

    def __call__(self, h5: h5tbx.File):
        """pivview post function. this is specific for a convention"""
        for name, ds in h5.items():
            if name in PIVVIEW_TRANSLATION:
                ds.attrs['standard_name', SSNO.standardName] = PIVVIEW_TRANSLATION[name]

        def _update_fields(grp: h5tbx.Group):
            piv_params = grp.attrs.get(PIV_PARAMETER_ATTRS_NAME, None)
            if piv_params is None:
                raise ValueError(f'No {PIV_PARAMETER_ATTRS_NAME} found in group {grp.name}')
            grp.attrs['piv_method'] = PIV_METHOD(
                piv_params['PIV processing parameters']['view0_piv_eval_method']).name
            grp.attrs['piv_peak_method'] = PIV_PeakFitMethod(
                piv_params['PIV processing parameters']['view0_piv_eval_peakfit_type']).name

            # for ds in grp.find({'$basename': {'$regex': 'planetime_[0-9]+'}}):
            #     ds.attrs['standard_name'] = pivview_translation['reltime']

        if PIV_PARAMETER_GRP_NAME not in h5:
            for param_grp in h5.find({'$basename': PIV_PARAMETER_GRP_NAME}, recursive=True, objfilter='group'):
                _update_fields(h5[param_grp.name])
            return
        return _update_fields(h5[PIV_PARAMETER_GRP_NAME])

        # # add `final_interrogation_window_size` to group "piv_parameters"
        # piv_parameter_groups = h5.find({'$basename': 'piv_parameters'})
        # if len(piv_parameter_groups) == 0:
        #     piv_parameter_groups = [h5.create_group('piv_parameters'),]
        #     # raise ValueError('No group named "piv_parameters" found in h5 file')
        # is_root_group = piv_parameter_groups[0].parent.name == '/'
        # if not is_root_group:
        #     piv_params = [_g.attrs.piv_parameters for _g in piv_parameter_groups]
        #     common_piv_params = _find_common_entries(piv_params)
        #     h5.create_group(PIV_PARAMETER_GRP_NAME)
        #     h5[PIV_PARAMETER_GRP_NAME].attrs['piv_parameters'] = common_piv_params
        #     for _g in piv_parameter_groups:
        #         _g.attrs['piv_parameters'] = {k: v for k, v in _g.attrs.piv_parameters.items() if k not in common_piv_params}
        #
        # piv_param_grp = h5[PIV_PARAMETER_GRP_NAME]
        #
        # piv_parameters = h5.attrs.piv_parameters
        # piv_proc_param = piv_parameters['PIV processing parameters']
        #
        # h5.attrs['piv_method'] = PIV_METHOD(int(piv_proc_param['view0_piv_eval_method'])).name
        #
        # if piv_proc_param['view0_piv_eval_method'] in (0, 1, 2):  # SinglePass and MultiPass
        #     x_final_iw_size, y_final_iw_size, _ = piv_proc_param['view0_piv_eval_samplesize']
        #     x_final_iw_overlap_size, y_final_iw_overlap_size, _ = piv_proc_param['view0_piv_eval_samplestep']
        # else:
        #     raise ValueError(f'Unknown PIV evaluation method: {piv_proc_param["view0_piv_eval_method"]}')
        # write_final_interrogation_window_size_to_h5_group(
        #     piv_param_grp, (x_final_iw_size, y_final_iw_size)
        # )
        # write_final_interrogation_overlap_size_to_h5_group(
        #     piv_param_grp,
        #     (x_final_iw_overlap_size, y_final_iw_overlap_size)
        # )


add_standard_name_operation = AddStandardNameOperation()



import logging
import unittest

from piv2hdf import set_loglevel, logger


class TestLogger(unittest.TestCase):

    def test_logger(self):

        for loglevel in (logging.CRITICAL, logging.ERROR, logging.WARN, logging.ERROR, logging.INFO):
            set_loglevel(loglevel)
            self.assertEqual(logger.level, loglevel)
            for handler in logger.handlers:
                self.assertEqual(handler.level, loglevel)



# """providing the save method that is not properly done in openPIV, at least until now (version 0.25.0)"""
#
# import numpy as np
# import pathlib
# from openpiv import filters, \
#     validation
# from typing import Dict, Union
#
#
# def outlier_detection(u, v,
#                       sig2noise,
#                       s2n_threshold=1.0,
#                       ulim=None,
#                       vlim=None,
#                       replace_outliers_kwargs: Union[Dict, bool] = None):
#     mask_s2n = validation.sig2noise_val(sig2noise,
#                                         threshold=s2n_threshold)  # only masks points where s2n is higher than threshold
#     if ulim is not None:
#         assert isinstance(vlim, (list, tuple))
#         assert len(ulim) == 2
#     if vlim is not None:
#         assert isinstance(vlim, (list, tuple))
#         assert len(ulim) == 2
#     if ulim is not None:
#         mask_vel_th_threshold = validation.global_val(u, v, ulim, vlim)
#         mask_outlier = mask_s2n | mask_vel_th_threshold
#     else:
#         mask_outlier = mask_s2n
#
#     # replace outliers in velocity field:
#     if replace_outliers_kwargs is not False:
#         _replace_outliers_kwargs = {'method': 'localmean', 'max_iter': 10, 'kernel_size': 5}
#         if replace_outliers_kwargs is not None:
#             _replace_outliers_kwargs.update(replace_outliers_kwargs)
#         u, v = filters.replace_outliers(u.copy(), v.copy(),
#                                         mask_outlier,
#                                         **_replace_outliers_kwargs)
#         interpolated = mask_outlier
#         disabled = None
#     else:
#         disabled = mask_outlier
#         interpolated = None
#
#     # flags =
#     # ACTIVE = 1  # not manipulated
#     # DISABLED = 2
#     # FILTERED = 4
#     # INTERPOLATED = 8
#     # REPLACED = 16
#     # MANUALEDIT = 32
#     # MASKED = 64
#
#     FLAGS = {'ACTIVE': 1,  # not manipulated
#              'DISABLED': 2,
#              'FILTERED': 4,
#              'INTERPOLATED': 8,
#              'REPLACED': 16,
#              'MANUALEDIT': 32,
#              'MASKED': 64}
#
#     mask = np.ones_like(u, dtype=int)
#     mask_dict = {  # 'ACTIVE': np.ones_like(x),
#         'DISABLED': disabled,
#         'FILTERED': None,
#         'INTERPOLATED': interpolated,
#         'REPLACED': None,
#         'MANUALEDIT': None,
#         'MASKED': None}
#
#     for k, m in mask_dict.items():
#         if m is not None:
#             mask[m] |= FLAGS[k]
#
#     return mask, u, v
#
#
# def save(
#         filename: Union[pathlib.Path, str],
#         x: np.ndarray,
#         y: np.ndarray,
#         u: np.ndarray,
#         v: np.ndarray,
#         *,
#         additional_data_fields: Dict,
#         fmt: str = "%.4e",
#         delimiter: str = "\t",
# ) -> None:
#     """Save flow field to an ascii file. This method in openpiv version 0.25.0
#     is not able to store other fields than x, y, u, v. This method is able to
#     store additional fields in the same file by accepting `additional_data_fields` (dict).
#
#     Parameters
#     ----------
#     filename : string
#         the path of the file where to save the flow field
#
#     x : 2d np.ndarray
#         a two dimensional array containing the x coordinates of the
#         interrogation window centers, in pixels.
#
#     y : 2d np.ndarray
#         a two dimensional array containing the y coordinates of the
#         interrogation window centers, in pixels.
#
#     u : 2d np.ndarray
#         a two dimensional array containing the u velocity components,
#         in pixels/seconds.
#
#     v : 2d np.ndarray
#         a two dimensional array containing the v velocity components,
#         in pixels/seconds.
#
#     additional_data_fields: Dict
#         Additional data arrays to save wiht same shape as x, y, u, v
#
#
#     fmt : string
#         a format string. See documentation of numpy.savetxt
#         for more details.
#
#     delimiter : string
#         character separating columns
#
#     Examples
#     --------
#
#     >>> piv2hdf.tools.save('field_001.txt', x, y, u, v, *,
#     >>>                    additional_data_fields={'mask': mask, 'flag': flag},  fmt='%6.3f',
#     >>>                    delimiter='\t')
#
#     """
#     base_shape = x.shape
#     for da in additional_data_fields.values():
#         if da.shape != base_shape:
#             raise ValueError(f"Shape of data array {da} is not equal to base shape {base_shape}")
#
#     if isinstance(u, np.ma.MaskedArray):
#         u = u.filled(0.)
#         v = v.filled(0.)
#
#     # build output array
#     out = np.vstack([m.flatten() for m in [x, y, u, v, *additional_data_fields.values()]])
#
#     header = f'{delimiter}'.join(['x', 'y', 'u', 'v', *additional_data_fields.keys()])
#
#     # header = "x"+ delimiter+ "y"+ delimiter+ "u"+ delimiter+ "v"+ delimiter+ "flags"+ delimiter+ "mask"
#     # save data to file.
#     np.savetxt(
#         filename,
#         out.T,
#         fmt=fmt,
#         delimiter=delimiter,
#         header=header,
#     )



from ast import literal_eval
from configparser import ConfigParser

import h5py

from .. import utils
from ..interface import PIVParameterInterface


class OpenPIVParameterFile(PIVParameterInterface):
    """open PIV parameter interface class"""

    __suffix__ = '.par'

    def __init__(self, filename):
        super().__init__(filename)
        if filename is not None:
            _cfg = ConfigParser()
            _cfg.read(filename)
            self.param_dict = {}
            if len(list(_cfg.sections())) == 1:
                for s in _cfg.sections():
                    self.param_dict = dict(_cfg[s])
            else:
                for s in _cfg.sections():
                    self.param_dict[s.strip('-').strip(' ')] = dict(_cfg[s])
            fwin = self._get_final_winsize()
            self.piv_param_datasets['x_final_iw_size']['data'] = fwin[0]
            self.piv_param_datasets['y_final_iw_size']['data'] = fwin[1]
            fov = self._get_overlap()
            self.piv_param_datasets['x_final_iw_overlap_size']['data'] = fov[0]
            self.piv_param_datasets['y_final_iw_overlap_size']['data'] = fov[1]
            self.piv_param_datasets['laser_pulse_delay']['data'] = float(self.param_dict['dt'])
            self.piv_param_datasets['piv_scaling_factor']['data'] = float(self.param_dict['scaling_factor'])

    def _get_final_winsize(self):
        """figure out the final interrogation window size"""
        winsize_data = self.param_dict['windowsizes']

        if isinstance(winsize_data, str):
            fwin = literal_eval(winsize_data)[-1]  # the last is the final interrogation window size
        else:
            fwin = winsize_data[-1]
        if isinstance(fwin, (tuple, list)):
            if len(fwin) == 2:
                fwin = [fwin[0], fwin[1], 1]
            elif len(fwin) == 3:
                fwin = [fwin[0], fwin[1], fwin[2]]
        else:
            fwin = [fwin, fwin, 1]
        return fwin

    def _get_overlap(self):
        overlap_size = self.param_dict['overlap']
        if isinstance(overlap_size, str):
            fov = literal_eval(overlap_size)[-1]
        else:
            fov = overlap_size[-1]
        if isinstance(fov, int):
            return [fov, fov, 1]
        # if isinstance(fov, (tuple, list)):
        if len(fov) == 2:
            return [fov[0], fov[1], 1]
        # elif len(fov) == 3:
        return [fov[0], fov[1], fov[2]]

    def save(self, filename):
        """Save parameter dictionary to file"""
        with open(filename, 'w') as f:
            f.write(f'[openpiv parameter]')
            for k, v in self.param_dict.items():
                line = f'\n{k}={v}'
                f.write(line.replace('%', '%%'))

    @staticmethod
    def from_windef(settings: "OpenPIVSetting"):
        """Initialize OpenPIVParameterFile from openpiv instance of class windef.Settings"""
        _param_dict = settings.__dict__.copy()
        try:
            _param_dict.pop('_FrozenClass__isfrozen')
        except KeyError:
            pass
        o = OpenPIVParameterFile(None)
        o.param_dict = _param_dict
        return o

    if True:
        def to_hdf(self, grp: h5py.Group):
            """Convert to HDF group"""

            def _to_grp(_dict, _grp):
                for k, v in _dict.items():
                    if isinstance(v, dict):
                        _grp = _to_grp(v, _grp.create_group(k))
                    else:
                        if k == 'windowsizes':
                            if isinstance(v, str):
                                fwin = literal_eval(v)[-1]  # the last is the final interrogation window size
                            else:
                                fwin = v[-1]
                            if isinstance(fwin, (tuple, list)):
                                if len(fwin) == 2:
                                    fwin = [fwin[0], fwin[1], 1]
                                elif len(fwin) == 3:
                                    fwin = [fwin[0], fwin[1], fwin[2]]
                            else:
                                fwin = [fwin, fwin, 1]

                            utils.write_final_interrogation_window_size_to_h5_group(_grp, fwin)
                        elif k == 'overlap':
                            if isinstance(v, str):
                                fov = literal_eval(v)[-1]
                            else:
                                fov = v[-1]
                            if isinstance(fov, (tuple, list)):
                                if len(fov) == 2:
                                    fov = [fov[0], fov[1], 1]
                                elif len(fov) == 3:
                                    fov = [fov[0], fov[1], fov[2]]
                            utils.write_final_interrogation_overlap_size_to_h5_group(_grp, fov)
                        elif k == 'dt':
                            ds = _grp.create_dataset(k, data=float(v))
                            ds.attrs.update({'units': 's', 'standard_name': 'laser_pulse_delay'})
                        elif k == 'scaling_factor':
                            ds = _grp.create_dataset(k, data=float(v))
                            ds.attrs.update({'units': 'pixel/m', 'standard_name': 'piv_scaling_factor'})
                        elif v is None:
                            _grp.attrs[k] = 'None'
                        else:
                            _grp.attrs[k] = v
                return _grp

            return _to_grp(self.param_dict, grp)



import pathlib

from h5rdmtoolbox import Convention

__this_dir__ = pathlib.Path(__file__).parent
RESOURCES_DIR = __this_dir__.parent / 'resources'
CONVENTION_FILENAME = RESOURCES_DIR / 'convention/standard_attributes.yaml'


def init_cv() -> Convention:
    """Initialize the convention from the YAML file."""
    if not CONVENTION_FILENAME.exists():
        raise ValueError(f'{CONVENTION_FILENAME} does not exist!')
    return Convention.from_yaml(CONVENTION_FILENAME, overwrite=True)


cv = init_cv()

if __name__ == '__main__':
    # force rewriting convention file
    cv = init_cv()
    import shutil

    shutil.rmtree(cv.filename.parent)
    cv = init_cv()



from pathlib import Path
from typing import Dict

import h5py
import numpy as np

try:
    from pyevtk.hl import gridToVTK
except ImportError:
    ImportError('Package pyevtk not installed. Either install it '
                'separately or install the repository with pip install piv2hdf [vtk]')

try:
    import vtk
    from vtk.util.numpy_support import numpy_to_vtk
except ImportError:
    ImportError('Package vtk not installed. Either install it '
                'separately or install the repository with pip install piv2hdf [vtk]')


def get_time_average_data_from_piv_case(hdf_input: Path or h5py.Group) -> Dict:
    """collects time-averaged data from a HDF PIV MULTI-PLANE and returns data as dictionary

    Parameters
    ----------
    hdf_input : Path or h5py.Group
        hdf filename or root group

    Returns
    --------
    data : Dict
    """

    def _collect(_h5, _data):
        _data['x'] = _h5['x'][()]  # vtk only accepts 3d data
        _data['y'] = _h5['y'][()]
        _data['z'] = _h5['z'][()]
        ta = _h5['compute_time_averages']
        for k in ta.keys():
            ds = ta[k]
            if ds.ndim == 4:
                for i in range(ds.shape[-1]):
                    _data[f'{ds.name}-{i}'] = ds[..., i].T
            else:
                _data[k] = ds[...].T
        return _data

    data = {}
    if isinstance(hdf_input, h5py.Group):
        return _collect(hdf_input, data)

    with h5py.File(hdf_input, 'r') as h5:
        data = _collect(h5, data)
    return data


def result_3D_to_vtk(data3: Dict, target_filename: str = None):
    """
    Saves 3D interpolated data in VTK regular grid format. Preconditioning to
    fortran style is handled in core function. Either to file or back to mayavi.

    Parameters
    ----------
    data3 : dict
        Input full 3D data as specified from above.
    target_filename : Path, optional=None
        If filepath given data is written to file.

    Returns
    -------
    vtkData : vtkRectilinearGrid
        The vtk gridded data for use in mayavi.
    file : 'str'
        vtk file name if passed

    Notes
    -----
    Credits to M. Elfner (ITS, Karlsruhe Institute of Technology)

    """
    pointDataRG, (xr, yr, zr) = precondition_grid_data(data3)

    save_to_file = True
    if target_filename is None:
        save_to_file = False

    file = None
    if save_to_file:
        file = gridToVTK(target_filename.__str__(), xr, yr, zr, pointData=pointDataRG)

    vtkData = numpy_data_to_rectgrid(xr, yr, zr, pointDataRG)

    return vtkData, file


def precondition_grid_data(datadict, invalidKeys=['x', 'y', 'z', 'info', 'nodes', 'transformAx']):
    """
    Preconditions numpy data for rectilinear grid in vtk. Now the rect grid is
    way way faster in post. Since paraview sucks so hard, everything is turned
    to fortran before saving...

    Parameters
    ----------
    datadict : dict
        Input dict with 3d PIV data, X, Y, Z are needed specifying the grid.
    invalidKeys : TYPE, optional
        Keys to remove from dict since they will kill paraview. The default is
        ['x', 'y', 'z', 'info', 'nodes'].

    Returns
    -------
    datadict : dict
        Dict with core reshaped data.
    tuple
        Coordinate arrays

    """
    xr, yr, zr = np.unique(datadict.pop('x')), np.unique(datadict.pop('y')), np.unique(datadict.pop('z'))
    # remove some invalid keys
    for s in invalidKeys:
        try:
            datadict.pop(s)
        except KeyError:
            ...
    # convert and save
    for k, v in datadict.items():
        if v.ndim > 3:
            datadict[k] = tuple(np.asfortranarray(v[..., i]).ravel('F') for i in range(v.shape[-1]))
        else:
            datadict[k] = np.asfortranarray(v).ravel('F')

    return datadict, (xr, yr, zr)


def numpy_data_to_rectgrid(x, y, z, data):
    """
    Creates a rectilinear dataset from numpy data as used in standard cases.
    DATA MUST BE PRECONDITIONED WITH FUNCTION ABOVE!!

    Parameters
    ----------
    x, y, z : np.ndarray
        1D point arrays specifyfing nodes on each axis.
    data : dict
        Dict with scalar or vector point data.

    Raises
    ------
    RuntimeError
        If data input problems occur.

    Returns
    -------
    rg : vtkRectilinearGrid
        The grid built from the numpy dataset.
    """

    def _pack(array, name):
        valid = False
        if isinstance(array, np.ndarray):
            if array.ndim == 1:
                dV = numpy_to_vtk(array, deep=True, array_type=vtk.VTK_FLOAT)
                dV.SetName(name)
                valid = True

        elif isinstance(array, tuple):
            if len(array) == 3:
                dV = np.asfortranarray(np.array(array).T)
                dV = numpy_to_vtk(dV, deep=True, array_type=vtk.VTK_FLOAT)
                dV.SetName(name)
                valid = True

        if not valid:
            raise RuntimeError('Cant convert data for vtk RG')

        return dV

    # Initialize
    rg = vtk.vtkRectilinearGrid()

    rg.SetDimensions(len(x), len(y), len(z))
    # set nodes
    rg.SetXCoordinates(numpy_to_vtk(x, deep=1, array_type=vtk.VTK_FLOAT))
    rg.SetYCoordinates(numpy_to_vtk(y, deep=1, array_type=vtk.VTK_FLOAT))
    rg.SetZCoordinates(numpy_to_vtk(z, deep=1, array_type=vtk.VTK_FLOAT))

    # Add data
    if isinstance(data, np.ndarray):
        _ = rg.GetPointData().AddArray(_pack(data, 'dfield_1'))

    elif isinstance(data, dict):
        for k, v in data.items():
            # if isinstance(v, tuple): continue
            _ = rg.GetPointData().AddArray(_pack(v, k))

    else:
        raise RuntimeError('Unsupported data type!')

    return rg



"""PIV FLAGS"""
import enum


# FLAGS = {'INACTIVE': 0,
#          'ACTIVE': 1,  # not manipulated
#          'DISABLED': 2,
#          'FILTERED': 4,
#          'INTERPOLATED': 8,
#          'REPLACED': 16,
#          'MANUALEDIT': 32,
#          'MASKED': 64}


class Flags(enum.Enum):
    """PIV validation flags. Adopted from PIVTec's PIVview."""
    INACTIVE = 0
    ACTIVE = 1
    MASKED = 2
    NORESULT = 4
    DISABLED = 8
    FILTERED = 16
    INTERPOLATED = 32
    REPLACED = 64
    MANUALEDIT = 128


# def get_flag_names(flag_value) -> List:
#     """Return a list of flag names for a given flag value"""
#     if flag_value == 0:
#         return ['INACTIVE']
#
#     flag_names = []
#
#     for flag in Flags:
#         if flag_value & flag.value:
#             flag_names.append(flag.name)
#
#     return flag_names


flag_translation_dict = {
    'openpiv': {
        0: Flags.ACTIVE,
        1: Flags.DISABLED,
        2: Flags.INTERPOLATED
    }
}
